# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2021 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************
# MIT License
#
# Copyright (c) 2018 Hunter Allen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import json
from typing import Any

import numpy as np


class NumpyEncoder(json.JSONEncoder):
    """Custom encoder for numpy data types"""

    def default(self, obj: Any) -> Any:
        if isinstance(
            obj,
            (
                np.int_,
                np.intc,
                np.intp,
                np.int8,
                np.int16,
                np.int32,
                np.int64,
                np.uint8,
                np.uint16,
                np.uint32,
                np.uint64,
            ),
        ):

            return int(obj)

        elif isinstance(obj, (float, np.float16, np.float32, np.float64)):
            return float(obj)

        elif isinstance(obj, (complex, np.complex64, np.complex128)):
            return {"real": obj.real, "imag": obj.imag}

        elif isinstance(obj, (np.ndarray,)):
            return obj.tolist()

        elif isinstance(obj, np.bool_):
            return bool(obj)

        elif isinstance(obj, np.void):
            return None

        return json.JSONEncoder.default(self, obj)



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2021 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

import asyncio
import logging
import traceback
from typing import Any, Optional, Tuple

import aiohttp

from .._logging import ErrorStore
from ..batch_utils import Batch, split_batch
from ..connection import send_asyncio_batch_requests, send_asyncio_is_alive
from ..exceptions import AsyncioSenderNoConnected
from ..logging_messages import (
    ASYNCIO_SENDER_HARD_SHUTDOWN,
    BATCH_SENDING_EXCEPTION,
    MPM_ALREADY_CLOSED_LOG_DATA_WARNING,
)
from ..server_address import ServerAddress
from ..utils import async_wait_for_done
from .base import CLOSE_MESSAGE, BaseSender

LOGGER = logging.getLogger(__name__)


def get_asyncio_sender(
    api_key: str,
    server_address: ServerAddress,
    max_batch_size: int,
    max_batch_time: int,
    batch_sending_timeout: int,
    error_store: Optional[ErrorStore] = None,
) -> "AsyncioSender":
    sender = AsyncioSender(
        api_key=api_key,
        server_address=server_address,
        max_batch_size=max_batch_size,
        max_batch_time=max_batch_time,
        batch_sending_timeout=batch_sending_timeout,
        error_store=error_store,
    )

    return sender


class AsyncioBackgroundSender:
    def __init__(
        self,
        api_key: str,
        server_address: ServerAddress,
        queue: "asyncio.Queue[Any]",
        max_batch_size: int,
        max_batch_time: int,
        batch_sending_timeout: int,
        error_store: Optional[ErrorStore] = None,
    ):
        """A background coroutine that reads from the queue, add the items to the batch and sends
        them whenever the batch is ready to be sent. Can only be created only when an event loop is running.
        """
        self.api_key = api_key
        self.server_address = server_address
        self.session: aiohttp.ClientSession = aiohttp.ClientSession()
        self.queue: "asyncio.Queue[Any]" = queue
        self.max_batch_time = max_batch_time
        self.batch = Batch(max_batch_size, self.max_batch_time)
        self.batch_sending_timeout = batch_sending_timeout
        self.error_store = error_store

        self.stop_processing = False

        # Pre-compute URLs
        self.batch_endpoint = self.server_address.batch_endpoint_url()
        self.labels_endpoint = self.server_address.batch_endpoint_url()

    async def run(self) -> None:
        while True:
            stop = await self._loop()

            if stop is True:
                break

        # We force stop the processing of the queue
        await self._check_batch_sending()
        return

    async def _loop(self) -> bool:
        try:
            data = await asyncio.wait_for(self.queue.get(), timeout=self.max_batch_time)
            if data is None:
                return False

            if data is CLOSE_MESSAGE:
                self.stop_processing = True
                return True

            self.batch.append(data)
        except asyncio.TimeoutError:
            pass

        await self._check_batch_sending()

        return False

    async def _send_batch(self, batch_to_send: Any) -> None:
        try:
            async with aiohttp.ClientSession() as session:
                predictions_batch, labels_batch = split_batch(batch_to_send)

                if predictions_batch:
                    await send_asyncio_batch_requests(
                        session=session,
                        url_endpoint=self.batch_endpoint,
                        api_key=self.api_key,
                        batch_sending_timeout=self.batch_sending_timeout,
                        batch=predictions_batch,
                        error_store=self.error_store,
                    )

                if labels_batch:
                    await send_asyncio_batch_requests(
                        session=session,
                        url_endpoint=self.labels_endpoint,
                        api_key=self.api_key,
                        batch_sending_timeout=self.batch_sending_timeout,
                        batch=labels_batch,
                        error_store=self.error_store,
                    )
        except Exception as e:
            # Log to console for immediate user visibility
            error_message = f"{BATCH_SENDING_EXCEPTION}. Error: {str(e)}"
            LOGGER.error(error_message)

            # Store in error store for programmatic access
            if self.error_store is not None:
                self.error_store.add_error(
                    message=error_message,
                    logger_name="comet_mpm.sender.asyncio_sender",
                    data_affected=batch_to_send,
                    traceback_info=traceback.format_exc(),
                )

    async def _check_batch_sending(self) -> None:
        # This should only be called by the background sender coroutine, if not the batch could
        # become corrupted as its access not protected by a lock
        try:
            if self.batch.should_be_uploaded(self.stop_processing):
                batch_to_send = self.batch.get_and_clear()

                predictions_batch, labels_batch = split_batch(batch_to_send)

                if predictions_batch:
                    await send_asyncio_batch_requests(
                        session=self.session,
                        url_endpoint=self.batch_endpoint,
                        api_key=self.api_key,
                        batch_sending_timeout=self.batch_sending_timeout,
                        batch=predictions_batch,
                        error_store=self.error_store,
                    )

                if labels_batch:
                    await send_asyncio_batch_requests(
                        session=self.session,
                        url_endpoint=self.labels_endpoint,
                        api_key=self.api_key,
                        batch_sending_timeout=self.batch_sending_timeout,
                        batch=labels_batch,
                        error_store=self.error_store,
                    )
        except Exception as e:
            # Log to console for immediate user visibility
            error_message = f"{BATCH_SENDING_EXCEPTION}. Error: {str(e)}"
            LOGGER.error(error_message)

            # Store in error store for programmatic access
            if self.error_store is not None:
                self.error_store.add_error(
                    message=error_message,
                    logger_name="comet_mpm.sender.asyncio_sender",
                    data_affected=batch_to_send,
                    traceback_info=traceback.format_exc(),
                )

    def close(self, timeout: int = 10) -> None:
        """Force the AsyncioBackgroundSender to stop processing messages"""
        #
        self.stop_processing = True

        async def flush_queue() -> None:
            while True:
                try:
                    data = self.queue.get_nowait()
                    if data is not None and data is not CLOSE_MESSAGE:
                        self.batch.append(data)
                except asyncio.QueueEmpty:
                    break

            batch = self.batch.get_and_clear()
            await asyncio.wait_for(self._send_batch(batch), timeout)
            await self.session.close()

        asyncio.run(flush_queue())


class AsyncioSender(BaseSender):
    def __init__(
        self,
        api_key: str,
        server_address: ServerAddress,
        max_batch_size: int,
        max_batch_time: int,
        batch_sending_timeout: int,
        error_store: Optional[ErrorStore] = None,
    ) -> None:
        self.sender_queue: Optional["asyncio.Queue[Any]"] = None
        # We need to wait for the asyncio event loop to be created
        self.background_sender: Optional[AsyncioBackgroundSender] = None
        self.background_task: Optional[asyncio.Task[None]] = None
        self.lock: Optional[asyncio.Lock] = None
        self.drain = False

        self.api_key = api_key
        self.server_address = server_address
        self.max_batch_time = max_batch_time
        self.max_batch_size = max_batch_size
        self.batch_sending_timeout = batch_sending_timeout
        self.error_store = error_store

    async def connect(self) -> None:
        if self.sender_queue is None:
            sender_queue: "asyncio.Queue[Any]" = asyncio.Queue()
            self.sender_queue = sender_queue
            self.background_sender = AsyncioBackgroundSender(
                api_key=self.api_key,
                server_address=self.server_address,
                queue=self.sender_queue,
                max_batch_size=self.max_batch_size,
                max_batch_time=self.max_batch_time,
                batch_sending_timeout=self.batch_sending_timeout,
                error_store=self.error_store,
            )
            self.background_sender.queue = self.sender_queue

        if self.background_task is None:
            assert self.background_sender is not None
            self.background_task = asyncio.create_task(self.background_sender.run())

        if self.lock is None:
            self.lock = asyncio.Lock()

        await self.ping_backend()

    async def ping_backend(self) -> None:
        assert self.background_sender is not None
        await send_asyncio_is_alive(
            session=self.background_sender.session,
            url_endpoint=self.server_address.is_alive_endpoint_url(),
            api_key=self.api_key,
        )

    def _check_connected(self) -> Tuple[asyncio.Lock, "asyncio.Queue[Any]"]:
        if self.lock is None or self.sender_queue is None:
            raise AsyncioSenderNoConnected()

        return self.lock, self.sender_queue

    async def put(self, item: Any) -> None:
        lock, queue = self._check_connected()
        async with lock:
            if not self.drain:
                await queue.put(item)
            else:
                LOGGER.error(MPM_ALREADY_CLOSED_LOG_DATA_WARNING)
                # Store warning for programmatic access
                if self.error_store is not None:
                    self.error_store.add_error(
                        message=MPM_ALREADY_CLOSED_LOG_DATA_WARNING,
                        logger_name="comet_mpm.sender.asyncio_sender",
                        data_affected=[item.create_event_dict()],
                        traceback_info=None,
                    )

    def close(self, timeout: int = 10) -> None:
        """Called during atexit but users should call join manually. In last resort, synchronously
        deque items and send it by recreating an asyncio Event Loop
        """
        if self.drain is True:
            return None
        else:
            self.drain = True

            need_to_clean = False

            if self.sender_queue is not None and self.sender_queue.qsize() > 0:
                need_to_clean = True

            if (
                self.background_sender is not None
                and len(self.background_sender.batch) > 0
            ):
                need_to_clean = True

            if need_to_clean:
                LOGGER.warning(
                    ASYNCIO_SENDER_HARD_SHUTDOWN
                )  # TODO: Add link to documentation page
                self.drain = True

                # If the sender_queue is not None, the background sender shouldn't be None either
                assert self.background_sender is not None
                self.background_sender.close(timeout)

    async def join(self, timeout: int) -> None:
        """Inform the background sender to stops processing after receiving the sentinel object and
        wait for the background sender to stop
        """
        try:
            lock, queue = self._check_connected()
        except AsyncioSenderNoConnected:
            # The AsyncioSender was not properly initialized, ignore
            return

        async with lock:
            if self.drain is False:
                self.drain = True

                queue.put_nowait(CLOSE_MESSAGE)

                if self.background_task is not None:

                    def is_task_done() -> bool:
                        if (
                            self.background_sender is not None
                            and self.background_task is not None
                        ):
                            return self.background_task.done()
                        return True

                    async def show_progress() -> None:
                        remaining_events = queue.qsize() if queue else 0
                        batch_size = (
                            len(self.background_sender.batch)
                            if self.background_sender
                            else 0
                        )
                        total_remaining = remaining_events + batch_size
                        LOGGER.info(
                            f"Uploading MPM events, {total_remaining} events remaining."
                        )

                    # Adjust sleep_time and progress callback based on timeout
                    # Always use responsive checking, but only show progress for longer timeouts
                    sleep_time = min(
                        10, max(timeout / 3, 1)
                    )  # Responsive checking for all timeouts
                    progress_callback = (
                        show_progress if timeout >= 15 else None
                    )  # Progress only for long timeouts

                    await async_wait_for_done(
                        check_function=is_task_done,
                        timeout=timeout,
                        progress_callback=progress_callback,
                        sleep_time=sleep_time,
                    )



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2021 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************
import logging
import queue
from threading import Lock, Thread
from typing import Any, Optional, Tuple

from .._logging import ErrorStore
from ..batch_utils import ThreadSafeBatch, split_batch
from ..connection import (
    get_comet_http_session,
    get_retry_strategy,
    send_batch_requests,
    send_is_alive,
)
from ..logging_messages import (
    BATCH_SENDING_EXCEPTION,
    MPM_ALREADY_CLOSED_LOG_DATA_WARNING,
)
from ..server_address import ServerAddress
from ..utils import wait_for_done
from .base import CLOSE_MESSAGE, BaseSender

LOGGER = logging.getLogger(__name__)


class ThreadBackgroundSender:
    def __init__(
        self,
        api_key: str,
        server_address: ServerAddress,
        sender_queue: "queue.Queue[Any]",
        max_batch_size: int,
        max_batch_time: int,
        batch_sending_timeout: int,
        error_store: Optional[ErrorStore] = None,
    ):
        self.api_key = api_key
        self.server_address = server_address
        self.sender_queue = sender_queue
        self.max_batch_time = max_batch_time
        self.batch = ThreadSafeBatch(max_batch_size, self.max_batch_time)
        self.session = get_comet_http_session(
            api_key=self.api_key, retry_strategy=get_retry_strategy()
        )
        self.batch_sending_timeout = batch_sending_timeout
        self.error_store = error_store

        self.stop_processing = False

        # Pre-compute URLs
        self.batch_endpoint = self.server_address.batch_endpoint_url()
        self.labels_endpoint = self.server_address.labels_endpoint_url()

    def run(self) -> None:
        """This is meant to be run in an independent Thread"""
        while self.stop_processing is False:
            stop = self._loop()

            if stop is True:
                break

        # We force stop the processing of the queue
        self._check_batch_sending()
        return

    def _loop(self) -> bool:
        try:
            data = self.sender_queue.get(block=True, timeout=self.max_batch_time)
            if data is None:
                return False

            if data is CLOSE_MESSAGE:
                self.stop_processing = True
                return True

            self.batch.append(data)
        except queue.Empty:
            pass

        self._check_batch_sending()

        return False

    def _check_batch_sending(self) -> None:
        try:
            if self.batch.should_be_uploaded(self.stop_processing):
                batch_to_send = self.batch.get_and_clear()

                predictions_batch, labels_batch = split_batch(batch_to_send)

                if predictions_batch:
                    send_batch_requests(
                        self.session,
                        self.batch_endpoint,
                        api_key=self.api_key,
                        batch=predictions_batch,
                        batch_sending_timeout=self.batch_sending_timeout,
                        error_store=self.error_store,
                    )

                if labels_batch:
                    send_batch_requests(
                        self.session,
                        self.labels_endpoint,
                        api_key=self.api_key,
                        batch=labels_batch,
                        batch_sending_timeout=self.batch_sending_timeout,
                        error_store=self.error_store,
                    )
        except Exception as e:
            # Log to console for immediate user visibility
            error_message = f"{BATCH_SENDING_EXCEPTION}. Error: {str(e)}"
            LOGGER.error(error_message)

            # Store in error store for programmatic access
            if self.error_store is not None:
                import traceback

                self.error_store.add_error(
                    message=error_message,
                    logger_name="comet_mpm.sender.thread_sender",
                    data_affected=None,
                    traceback_info=traceback.format_exc(),
                )

    def close(self) -> None:
        """For the BackgroundSender to stop processing messages"""
        self.stop_processing = True
        # This shouldn't happen outside of the background sender thread
        self._check_batch_sending()

        self.session.close()


class ThreadSender(BaseSender):
    def __init__(
        self,
        api_key: str,
        server_address: ServerAddress,
        max_batch_size: int,
        max_batch_time: int,
        batch_sending_timeout: int,
        error_store: Optional[ErrorStore] = None,
    ) -> None:
        self.lock = Lock()
        self.sender_queue: Optional["queue.Queue[Any]"] = None
        self.background_sender: Optional[ThreadBackgroundSender] = None
        self.background_thread: Optional[Thread] = None
        self.drain = False

        self.api_key = api_key
        self.server_address = server_address
        self.max_batch_time = max_batch_time
        self.max_batch_size = max_batch_size
        self.batch_sending_timeout = batch_sending_timeout
        self.error_store = error_store

    def _prepare(self) -> "queue.Queue[Any]":
        """We need to create the background thread on the first request to support application
        pre-loading.
        """
        if self.sender_queue is None:
            sender_queue: "queue.Queue[Any]" = queue.Queue()
            self.sender_queue = sender_queue
            self.background_sender = ThreadBackgroundSender(
                api_key=self.api_key,
                server_address=self.server_address,
                sender_queue=self.sender_queue,
                max_batch_size=self.max_batch_size,
                max_batch_time=self.max_batch_time,
                batch_sending_timeout=self.batch_sending_timeout,
                error_store=self.error_store,
            )

        if self.background_thread is None:
            assert self.background_sender is not None
            self.background_thread = Thread(
                target=self.background_sender.run, daemon=True, name="ThreadSender"
            )
            self.background_thread.start()

        assert self.sender_queue is not None
        return self.sender_queue

    def put(self, item: Any) -> None:
        with self.lock:
            if not self.drain:

                if self.sender_queue is None:
                    sender_queue = self._prepare()
                else:
                    sender_queue = self.sender_queue

                sender_queue.put(item)
            else:
                LOGGER.error(MPM_ALREADY_CLOSED_LOG_DATA_WARNING)
                # Store warning for programmatic access
                if self.error_store is not None:
                    self.error_store.add_error(
                        message=MPM_ALREADY_CLOSED_LOG_DATA_WARNING,
                        logger_name="comet_mpm.sender.thread_sender",
                        data_affected=[item.create_event_dict()],
                        traceback_info=None,
                    )

    def close(self, timeout: int = 10) -> None:
        with self.lock:
            self.drain = True

        if self.sender_queue is not None:
            self.sender_queue.put(CLOSE_MESSAGE)

        if self.background_thread is not None:

            def is_thread_done() -> bool:
                if (
                    self.background_sender is not None
                    and self.background_thread is not None
                ):
                    return not self.background_thread.is_alive()
                return True

            def show_progress() -> None:
                remaining_events = self.sender_queue.qsize() if self.sender_queue else 0
                LOGGER.info(
                    f"Uploading MPM events, {remaining_events} events remaining."
                )

            # Adjust sleep_time and progress callback based on timeout
            # Always use responsive checking, but only show progress for longer timeouts
            sleep_time = min(
                10, max(timeout / 3, 1)
            )  # Responsive checking for all timeouts
            progress_callback = (
                show_progress if timeout >= 15 else None
            )  # Progress only for long timeouts

            wait_for_done(
                check_function=is_thread_done,
                timeout=timeout,
                progress_callback=progress_callback,
                sleep_time=sleep_time,
            )

        if self.background_sender is not None:
            self.background_sender.close()

    def join(self, timeout: int) -> None:
        """There is no easy way to plug a shutdown callback with Flask, the "normal" cleaning
        process for Flask is to use the close method"""
        return None

    def connect(self) -> None:
        """There is no easy way nor need to plug a startup callback as we can create everything and
        do the handshake during Python import"""
        self.ping_backend()

    def ping_backend(self) -> None:
        if self.background_sender is None:
            session = get_comet_http_session(
                api_key=self.api_key, retry_strategy=get_retry_strategy()
            )
        else:
            session = self.background_sender.session
        send_is_alive(
            session=session,
            url_endpoint=self.server_address.is_alive_endpoint_url(),
            api_key=self.api_key,
        )


def get_thread_sender(
    api_key: str,
    server_address: ServerAddress,
    max_batch_size: int,
    max_batch_time: int,
    batch_sending_timeout: int,
    error_store: Optional[ErrorStore] = None,
) -> "ThreadSender":
    sender = ThreadSender(
        api_key=api_key,
        server_address=server_address,
        max_batch_size=max_batch_size,
        max_batch_time=max_batch_time,
        batch_sending_timeout=batch_sending_timeout,
        error_store=error_store,
    )

    return sender


def _get_thread_background_sender(
    api_key: str,
    server_address: ServerAddress,
    max_batch_size: int,
    max_batch_time: int,
    batch_sending_timeout: int,
) -> Tuple["queue.Queue[Any]", "ThreadBackgroundSender"]:
    sender_queue: queue.Queue[Any] = queue.Queue()
    background_sender = ThreadBackgroundSender(
        api_key=api_key,
        server_address=server_address,
        sender_queue=sender_queue,
        max_batch_size=max_batch_size,
        max_batch_time=max_batch_time,
        batch_sending_timeout=batch_sending_timeout,
    )

    return sender_queue, background_sender



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2025 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************
import asyncio
import sys
import time
from typing import Awaitable, Callable, Optional

if sys.version_info[:2] >= (3, 8):
    import importlib.metadata as importlib_metadata
else:
    import importlib_metadata


def get_comet_version() -> str:
    try:
        return importlib_metadata.version("comet_mpm")
    except importlib_metadata.PackageNotFoundError:
        return "Please install comet with `pip install comet_mpm`"


def wait_for_done(
    check_function: Callable[[], bool],
    timeout: float,
    progress_callback: Optional[Callable[[], None]] = None,
    sleep_time: float = 1,
) -> None:
    """Wait up to timeout seconds for the check function to return True.

    Args:
        check_function: Function that returns True when condition is met
        timeout: Maximum time to wait in seconds
        progress_callback: Optional function called periodically during wait
        sleep_time: Interval between progress callback calls in seconds
    """
    end_time = time.time() + timeout
    while check_function() is False and time.time() < end_time:
        if progress_callback is not None:
            progress_callback()
        # Wait a max of sleep_time, but keep checking to see if
        # check_function is done. Allows wait_for_done to end
        # before sleep_time has elapsed:
        end_sleep_time = time.time() + sleep_time
        while check_function() is False and time.time() < end_sleep_time:
            time.sleep(sleep_time / 20.0)


async def async_wait_for_done(
    check_function: Callable[[], bool],
    timeout: float,
    progress_callback: Optional[Callable[[], Awaitable[None]]] = None,
    sleep_time: float = 1,
) -> None:
    """Async version: Wait up to timeout seconds for the check function to return True.

    Args:
        check_function: Function that returns True when condition is met
        timeout: Maximum time to wait in seconds
        progress_callback: Optional async function called periodically during wait
        sleep_time: Interval between progress callback calls in seconds
    """
    end_time = time.time() + timeout
    while check_function() is False and time.time() < end_time:
        if progress_callback is not None:
            await progress_callback()
        # Wait a max of sleep_time, but keep checking to see if
        # check_function is done. Allows async_wait_for_done to end
        # before sleep_time has elapsed:
        end_sleep_time = time.time() + sleep_time
        while check_function() is False and time.time() < end_sleep_time:
            await asyncio.sleep(sleep_time / 20.0)



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2021-2025 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

from typing import Optional


def remove_prefix(name: str) -> str:
    """
    Remove the prefix from a prefixed name.

    Args:
        name: A string with a prefix separated by underscore (e.g., "feature_age")

    Returns:
        str: The name without the prefix (e.g., "age")

    Raises:
        ValueError: If the name doesn't contain an underscore
    """
    prefix, rest = name.split("_", 1)
    return rest


def source_type_to_prefix(source_type: Optional[str]) -> str:
    """
    Convert a source type to its corresponding prefix.

    Args:
        source_type: The source type string or None

    Returns:
        str: The corresponding prefix ("label", "prediction", or "feature")

    Raises:
        Exception: If the source type is unknown
    """
    if source_type is None:
        return "label"
    elif source_type == "model_output_features":
        return "prediction"
    elif source_type == "model_input_features":
        return "feature"
    else:
        raise ValueError("unknown source type: %r" % source_type)



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2021-2025 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

import inspect
import os
from abc import ABC, abstractmethod
from datetime import datetime, timedelta
from typing import Any, Dict, Optional


def find_page_name() -> Optional[str]:
    """
    This function searches the call stack to find the name of the calling
    Python Panel function. It depends on the name of the panel script.
    """
    stack = inspect.stack()
    for row in stack:
        frame, filename, line_number, function_name, lines, index = row
        basename = os.path.basename(filename)
        # NOTE: this is based on specific naming of the Python Panel files:
        if basename.startswith("edit_mpm_panel_") or basename.startswith("mpm_panel_"):
            return basename

    return None


class ConfigStrategy(ABC):
    """Abstract base class for configuration strategies."""

    @abstractmethod
    def get(self, key: str) -> Any:
        """Get a configuration value."""
        pass

    @abstractmethod
    def get_option(self, option_key: str) -> Any:
        """Get a specific option with proper fallback."""
        pass


class PythonPanelConfig(ConfigStrategy):
    """Configuration strategy for Python Panel environment."""

    def __init__(self, page_name: str, config_override: Dict[str, Any]):
        """
        Initialize with page name and config override.

        Args:
            page_name: Name of the panel page
            config_override: Configuration override dictionary
        """
        self._page_name = page_name
        self._config_override = config_override

    def get(self, key: str) -> Any:
        """
        Get the key in the config from Python Panel environment.
        """
        config = self._config_override.get(self._page_name, {})
        return config.get(key, None)

    def get_option(self, option_key: str) -> Any:
        """
        Get a specific option from COMET_PANEL_OPTIONS with proper fallback.

        This method ensures that even if the panel options are overridden
        and don't contain a specific key, we fall back to the default values.

        Args:
            option_key: The specific option key to retrieve

        Returns:
            The option value, or None if not found in either source
        """
        # First try to get from the current panel options
        panel_options = self.get("COMET_PANEL_OPTIONS")
        if panel_options is not None:
            value = panel_options.get(option_key)
            if value is not None:
                return value

        # If not found or panel_options is None, get the default
        default_options = self._get_default_panel_options()
        if default_options is not None:
            return default_options.get(option_key)

        return None

    def _get_default_panel_options(self) -> Optional[Dict[str, Any]]:
        """Get the default panel options."""
        # Get current time in local timezone
        now = datetime.now()
        # Calculate 30 days ago
        thirty_days_ago = now - timedelta(days=30)

        return {
            "filters": [],
            "intervalType": "DAILY",
            "startDate": thirty_days_ago.strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3] + "Z",
            "endDate": now.strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3] + "Z",
        }


class DefaultConfig(ConfigStrategy):
    """Configuration strategy for hard-coded defaults."""

    def get(self, key: str) -> Any:
        """
        Get configuration values from hard-coded defaults.
        """
        if key == "COMET_PANEL_OPTIONS":
            # Get current time in local timezone
            now = datetime.now()
            # Calculate 30 days ago
            thirty_days_ago = now - timedelta(days=30)

            return {
                "filters": [],
                "intervalType": "DAILY",
                "startDate": thirty_days_ago.strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3]
                + "Z",
                "endDate": now.strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3] + "Z",
            }

        elif key == "comet.url_override":
            return os.environ.get("COMET_URL_OVERRIDE")

        elif key == "comet.api_key":
            return os.environ.get("COMET_API_KEY")

        elif key == "comet.workspace":
            return os.environ.get("COMET_WORKSPACE")

        elif key == "COMET_PANEL_WIDTH":
            return 901

        elif key == "COMET_PANEL_HEIGHT":
            return 248

        else:
            return None

    def get_option(self, option_key: str) -> Any:
        """
        Get a specific option from COMET_PANEL_OPTIONS.

        Args:
            option_key: The specific option key to retrieve

        Returns:
            The option value, or None if not found
        """
        panel_options = self.get("COMET_PANEL_OPTIONS")
        if panel_options is not None:
            return panel_options.get(option_key)
        return None


def get_config() -> ConfigStrategy:
    """
    Factory function that returns the appropriate config strategy.

    Returns:
        ConfigStrategy: The appropriate configuration strategy based on the environment
    """
    try:
        import streamlit as st
    except ImportError:
        return DefaultConfig()

    page_name = find_page_name()
    if page_name is None:
        return DefaultConfig()

    config_override = st.session_state.get("comet_config_override")
    if not isinstance(config_override, dict):
        return DefaultConfig()

    return PythonPanelConfig(page_name, config_override)



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2021 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

from typing import Dict, Optional, Tuple


def get_request_data(
    base_url: Optional[str],
    workspace_name: str,
    model_name: str,
    model_version: str,
    dataset_type: str,
    dataset_name: Optional[str] = None,
    na_values: Optional[str] = None,
    keep_default_na_values: Optional[str] = None,
) -> Tuple[Dict[str, str], str]:
    data = {
        "workspaceName": workspace_name,
        "modelName": model_name,
        "modelVersion": model_version,
    }
    url = set_model_name_and_get_url(base_url, data, dataset_type, dataset_name)
    if na_values is not None:
        data["NAValues"] = na_values
    if keep_default_na_values is not None:
        data["keepDefaultNAValues"] = keep_default_na_values
    return data, url


def set_model_name_and_get_url(
    base_url: Optional[str],
    data: Dict[str, str],
    dataset_type: str,
    dataset_name: Optional[str] = None,
) -> str:
    if dataset_type == "TRAINING_EVENTS":
        if dataset_name is None:
            raise ValueError(
                "dataset_name must be set when dataset_type is 'TRAINING_EVENTS'."
            )
        data["datasetName"] = dataset_name
        url = f"{base_url}v2/dataset/events/csv"
    elif dataset_type == "EVENTS":
        if dataset_name is not None:
            data["modelName"] = dataset_name
        url = f"{base_url}v2/data/events/csv"
    elif dataset_type == "LATE_LABELS":
        if dataset_name is not None:
            data["modelName"] = dataset_name
        url = f"{base_url}v2/data/labels/csv"
    else:
        raise ValueError(
            "Invalid dataset_type. Allowed values are 'EVENTS', 'LATE_LABELS' or 'TRAINING_EVENTS'."
        )
    return url



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2021-2024 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

import logging
from typing import Optional, Union

from .api_key.comet_api_key import parse_api_key
from .connection_helpers import get_root_url, sanitize_url
from .logging_messages import BASE_URL_MISMATCH_CONFIG_API_KEY
from .settings import AnyHttpUrl, MPMSettings

LOGGER = logging.getLogger(__name__)
DEFAULT_COMET_BASE_URL = "https://www.comet.com/"


def extract_comet_url(settings: MPMSettings) -> str:
    """
    Extracts Comet base url from settings or API key and sanitizes it (appends / at the end)
    """
    return get_comet_base_url(settings.api_key, settings.url)


def get_comet_base_url(
    settings_api_key: Optional[str], settings_url: Optional[Union[str, AnyHttpUrl]]
) -> str:
    """
    Determines the Comet base URL based on the provided API key and URL.

    Args:
        settings_api_key (Optional[str]): The API key for Comet, which may contain a base URL.
        settings_url (Optional[Union[str, AnyHttpUrl]]): The URL explicitly set in the settings.

    Returns:
        str: The sanitized Comet base URL, with a trailing slash if not already present.
    """
    if settings_api_key is None:
        return DEFAULT_COMET_BASE_URL

    api_key = parse_api_key(settings_api_key)
    if api_key is None:
        return DEFAULT_COMET_BASE_URL

    if settings_url is not None:
        settings_base_url = sanitize_url(get_root_url(str(settings_url)))
        if (
            api_key.base_url is not None
            and sanitize_url(api_key.base_url) != settings_base_url
        ):
            LOGGER.warning(
                BASE_URL_MISMATCH_CONFIG_API_KEY,
                settings_base_url,
                sanitize_url(api_key.base_url),
            )
        # do not change base url but add trailing slash if not there
        return sanitize_url(str(settings_url))

    if api_key.base_url is not None:
        return sanitize_url(api_key.base_url)

    return DEFAULT_COMET_BASE_URL



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.com
#  Copyright (C) 2021-2024 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************
import logging

from .aws_lambda.lambda_helpers import is_aws_lambda_environment
from .logging_messages import MPM_IN_AWS_LAMBDA_NEEDS_END

LOGGER = logging.getLogger(__name__)


def check_environment() -> None:
    if is_aws_lambda_environment():
        LOGGER.warning(MPM_IN_AWS_LAMBDA_NEEDS_END)



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2021 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************
import logging
from typing import Any, Dict, List, Optional

from .. import constants, logging_messages
from .prediction_event import PredictionEvent

LOGGER = logging.getLogger(__name__)


def generate(  # type: ignore[no-untyped-def]
    workspace: str,
    model_name: str,
    model_version: str,
    dataframe,
    prediction_id_column: str,
    feature_columns: Optional[List[str]] = None,
    output_features_columns: Optional[List[str]] = None,
    output_value_column: Optional[str] = None,
    output_probability_column: Optional[str] = None,
    labels_columns: Optional[List[str]] = None,
    timestamp_column: Optional[str] = None,
):
    return (
        _event(
            workspace=workspace,
            model_name=model_name,
            model_version=model_version,
            row=row,
            prediction_id_column=prediction_id_column,
            feature_columns=feature_columns,
            output_features_columns=output_features_columns,
            output_value_column=output_value_column,
            output_probability_column=output_probability_column,
            labels_columns=labels_columns,
            timestamp_column=timestamp_column,
        )
        for row in dataframe.to_dict(orient="records")
    )


def _event(  # type: ignore[no-untyped-def]
    workspace: str,
    model_name: str,
    model_version: str,
    row,
    prediction_id_column,
    feature_columns,
    output_features_columns,
    output_value_column,
    output_probability_column,
    labels_columns: Optional[List[str]] = None,
    timestamp_column: Optional[str] = None,
) -> PredictionEvent:
    prediction_id = str(row[prediction_id_column])
    input_features = None
    if feature_columns is not None:
        input_features = {key: row[key] for key in feature_columns}

    output_value = None
    if output_value_column is not None:
        output_value = row[output_value_column]

    output_probability = None
    if output_probability_column is not None:
        output_probability = row[output_probability_column]

    output_features = None
    if output_features_columns is not None:
        output_features = {key: row[key] for key in output_features_columns}
    output_features = _handle_dataframe_output_features(
        output_value, output_probability, output_features
    )

    labels = None
    if labels_columns is not None:
        labels = {key: row[key] for key in labels_columns}

    timestamp = None
    if timestamp_column is not None:
        timestamp = row[timestamp_column]

    return PredictionEvent(
        workspace=workspace,
        model_name=model_name,
        model_version=model_version,
        prediction_id=prediction_id,
        input_features=input_features,
        output_features=output_features,
        labels=labels,
        timestamp=timestamp,
    )


def _handle_dataframe_output_features(
    output_value: Any,
    output_probability: Any,
    output_features: Optional[Dict[str, Any]],
) -> Optional[Dict[str, Any]]:
    event_output_features: Optional[Dict[str, Any]]

    if (
        output_value is not None or output_probability is not None
    ) and output_features is None:
        LOGGER.warning(
            logging_messages.DEPRECATED_DATAFRAME_OUTPUT_VALUE_AND_PROBABILITY
        )
        event_output_features = {}
        if output_value is not None:
            event_output_features[constants.EVENT_PREDICTION_VALUE] = output_value

        if output_probability is not None:
            event_output_features[
                constants.EVENT_PREDICTION_PROBABILITY
            ] = output_probability
    elif (
        output_value is not None and output_probability is not None
    ) and output_features is not None:
        LOGGER.warning(
            logging_messages.DEPRECATED_DATAFRAME_OUTPUT_VALUE_AND_PROBABILITY
        )

        event_output_features = {
            constants.EVENT_PREDICTION_VALUE: output_value,
            constants.EVENT_PREDICTION_PROBABILITY: output_probability,
        }
        for key in output_features:
            event_output_features[key] = output_features[key]
    else:
        event_output_features = output_features

    return event_output_features



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2021-2025 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

"""
API objects for Comet MPM.

This package provides high-level API objects for interacting with the Comet MPM
backend, including models, REST clients, and data converters.
"""

from .api import API
from .model.model import Model

__all__ = ["API", "Model"]



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2021-2025 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

from typing import Any, Dict, List, Optional

import pandas as pd

from ...config import get_config
from ..rest_client import Client
from .dataframe_converters import (
    custom_metrics_to_dataframe,
    feature_category_distribution_to_dataframe,
    feature_density_to_dataframe,
    feature_drift_to_dataframe,
    feature_percentiles_to_dataframe,
)


def convert_config_filters_to_predicates(filters: List[str]) -> List[Dict[str, str]]:
    """
    Convert a list of filter strings to predicate format.

    Args:
        filters: List of filter strings to convert

    Returns:
        List[Dict[str, str]]: List of predicate dictionaries with "key" and "query" fields
    """
    return [{"key": filter, "query": filter} for filter in filters]


class Model:
    """
    A model instance for interacting with Comet MPM model-specific operations.

    This class provides high-level methods for querying model predictions, metrics,
    and feature analysis. It can be configured with panel options for default
    parameter values.

    Args:
        client: The Comet MPM client instance
        model_id: The ID of the model to work with
    """

    def __init__(
        self,
        client: Client,
        model_id: str,
    ):
        """
        Initialize a Model instance.

        Args:
            client: The Comet MPM client instance for making API calls
            model_id: The ID of the model to work with
            _config: Configuration strategy instance
        """
        self._client = client
        self.model_id = model_id
        self._config = get_config()

    def get_details(self) -> Dict[str, Any]:
        """
        Get the details of a model.

        Returns:
            Dict[str, Any]: Model details including metadata, configuration, and status
        """
        return self._client.get_model_details(self.model_id)

    def get_nb_predictions(
        self,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
        interval_type: Optional[str] = None,
        filters: Optional[List[str]] = None,
        model_version: Optional[str] = None,
    ) -> pd.DataFrame:
        """
        Get the number of predictions for a model within a specified time range.

        Args:
            start_date: Start date for filtering predictions (ISO format)
            end_date: End date for filtering predictions (ISO format)
            interval_type: Type of interval for aggregation ("DAILY" or "HOURLY")
            filters: List of filters to apply to predictions
            model_version: Specific model version to query

        Returns:
            pd.DataFrame: DataFrame containing the number of predictions matching the criteria
        """
        # Use SQL for now:
        df = self.get_custom_metric(
            "SELECT count(*) FROM model",
            start_date=start_date,
            end_date=end_date,
            interval_type=interval_type,
            filters=filters,
            model_version=model_version,
        )
        return df

    def get_custom_metric(
        self,
        sql: str,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
        interval_type: Optional[str] = None,
        filters: Optional[List[str]] = None,
        model_version: Optional[str] = None,
    ) -> pd.DataFrame:
        """
        Execute a custom SQL query to retrieve model metrics.

        Args:
            sql: SQL query string to execute
            start_date: Start date for filtering results (ISO format)
            end_date: End date for filtering results (ISO format)
            interval_type: Type of interval for aggregation ("DAILY" or "HOURLY")
            filters: List of filters to apply to results
            model_version: Specific model version to query

        Returns:
            DataFrame: Results of the SQL query
        """
        if start_date is None:
            start_date = self._config.get_option("startDate")
        if end_date is None:
            end_date = self._config.get_option("endDate")
        if interval_type is None:
            interval_type = self._config.get_option("intervalType")
        if filters is None:
            filters = self._config.get_option("filters")
        if model_version is None:
            model_version = self._config.get_option("modelVersion")

        predicates = convert_config_filters_to_predicates(filters or [])
        data = self._client.get_custom_metrics(
            model_id=self.model_id,
            sql=sql,
            start_date=start_date,
            end_date=end_date,
            interval_type=interval_type,
            predicates=predicates,
            model_version=model_version,
        )

        df = custom_metrics_to_dataframe(data, sql)
        df.attrs.update(
            {
                "method": "get_custom_metrics",
                "sql": sql,
                "start_date": start_date,
                "end_date": end_date,
                "interval_type": interval_type,
                "filters": filters,
                "model_version": model_version,
            }
        )
        return df

    def get_feature_drift(
        self,
        feature_name: str,
        algorithm: str = "EMD",
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
        interval_type: Optional[str] = None,
        filters: Optional[List[str]] = None,
        model_version: Optional[str] = None,
    ) -> pd.DataFrame:
        """
        Calculate drift metrics for a specific feature.

        Args:
            feature_name: Name of the feature to calculate drift for
            algorithm: Drift calculation algorithm ("EMD", "PSI", or "KL")
            start_date: Start date for drift calculation (ISO format)
            end_date: End date for drift calculation (ISO format)
            interval_type: Type of interval for aggregation ("DAILY" or "HOURLY")
            filters: List of filters to apply to drift calculation
            model_version: Specific model version to query

        Returns:
            DataFrame: Drift metrics for the specified feature
        """
        if start_date is None:
            start_date = self._config.get_option("startDate")
        if end_date is None:
            end_date = self._config.get_option("endDate")
        if interval_type is None:
            interval_type = self._config.get_option("intervalType")
        if filters is None:
            filters = self._config.get_option("filters")
        if model_version is None:
            model_version = self._config.get_option("modelVersion")

        predicates = convert_config_filters_to_predicates(filters or [])
        data = self._client.get_feature_drift(
            feature_name=feature_name,
            algorithm=algorithm,
            model_id=self.model_id,
            start_date=start_date,
            end_date=end_date,
            interval_type=interval_type,
            predicates=predicates,
            model_version=model_version,
        )
        df = feature_drift_to_dataframe(data)
        df.attrs.update(
            {
                "method": "get_feature_drift",
                "feature_name": feature_name,
                "start_date": start_date,
                "end_date": end_date,
                "interval_type": interval_type,
                "filters": filters,
                "model_version": model_version,
            }
        )
        return df

    def get_feature_category_distribution(
        self,
        feature_name: str,
        normalize: bool = False,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
        interval_type: Optional[str] = None,
        filters: Optional[List[str]] = None,
        model_version: Optional[str] = None,
    ) -> pd.DataFrame:
        """
        Get the distribution of categories for a categorical feature.

        Args:
            feature_name: Name of the categorical feature
            normalize: If True, returns percentages instead of counts
            start_date: Start date for distribution calculation (ISO format)
            end_date: End date for distribution calculation (ISO format)
            interval_type: Type of interval for aggregation ("DAILY" or "HOURLY")
            filters: List of filters to apply to distribution calculation
            model_version: Specific model version to query

        Returns:
            DataFrame: Distribution of feature categories
        """
        if start_date is None:
            start_date = self._config.get_option("startDate")
        if end_date is None:
            end_date = self._config.get_option("endDate")
        if interval_type is None:
            interval_type = self._config.get_option("intervalType")
        if filters is None:
            filters = self._config.get_option("filters")
        if model_version is None:
            model_version = self._config.get_option("modelVersion")

        predicates = convert_config_filters_to_predicates(filters or [])
        data = self._client.get_feature_category_distribution(
            feature_name=feature_name,
            model_id=self.model_id,
            normalize=normalize,
            start_date=start_date,
            end_date=end_date,
            interval_type=interval_type,
            predicates=predicates,
            model_version=model_version,
        )
        df = feature_category_distribution_to_dataframe(data)
        df.attrs.update(
            {
                "method": "get_feature_category_distribution",
                "feature_name": feature_name,
                "start_date": start_date,
                "end_date": end_date,
                "interval_type": interval_type,
                "filters": filters,
                "model_version": model_version,
            }
        )
        return df

    def get_feature_density(
        self,
        feature_name: str,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
        filters: Optional[List[str]] = None,
        model_version: Optional[str] = None,
        interval_type: Optional[str] = None,
    ) -> pd.DataFrame:
        """
        Get the probability density function (PDF) of a numeric feature.

        Args:
            feature_name: Name of the numeric feature
            start_date: Start date for density calculation (ISO format)
            end_date: End date for density calculation (ISO format)
            filters: List of filters to apply to density calculation
            model_version: Specific model version to query
            interval_type: Type of interval for aggregation ("DAILY" or "HOURLY")

        Returns:
            DataFrame: Probability density function of the feature values
        """
        if start_date is None:
            start_date = self._config.get_option("startDate")
        if end_date is None:
            end_date = self._config.get_option("endDate")
        if filters is None:
            filters = self._config.get_option("filters")
        if model_version is None:
            model_version = self._config.get_option("modelVersion")
        if interval_type is None:
            interval_type = self._config.get_option("intervalType")

        predicates = convert_config_filters_to_predicates(filters or [])
        data = self._client.get_feature_density(
            model_id=self.model_id,
            feature_name=feature_name,
            start_date=start_date,
            end_date=end_date,
            predicates=predicates,
            model_version=model_version,
            interval_type=interval_type,
        )
        df = feature_density_to_dataframe(data)
        df.attrs.update(
            {
                "method": "get_feature_density",
                "feature_name": feature_name,
                "start_date": start_date,
                "end_date": end_date,
                "interval_type": interval_type,
                "filters": filters,
                "model_version": model_version,
            }
        )
        return df

    def get_feature_percentiles(
        self,
        feature_name: str,
        percentiles: Optional[List[float]] = None,  # Only these are supported
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
        filters: Optional[List[str]] = None,
        model_version: Optional[str] = None,
        interval_type: Optional[str] = None,
    ) -> pd.DataFrame:
        """
        Get the specified percentiles for a numeric feature.

        Args:
            feature_name: Name of the numeric feature
            percentiles: List of percentiles to calculate (default: [0, 0.1, 0.25, 0.5, 0.75, 0.9, 1])
                Only these specific percentile values are supported
            start_date: Start date for percentile calculation (ISO format)
            end_date: End date for percentile calculation (ISO format)
            filters: List of filters to apply to percentile calculation
            model_version: Specific model version to query
            interval_type: Type of interval for aggregation. Must be one of "DAILY" or "HOURLY".
                Passing any other value will raise a ValueError.

        Returns:
            DataFrame: Percentile values for the specified feature
        """
        if percentiles is None:
            percentiles = [0, 0.1, 0.25, 0.5, 0.75, 0.9, 1]
        if start_date is None:
            start_date = self._config.get_option("startDate")
        if end_date is None:
            end_date = self._config.get_option("endDate")
        if interval_type is None:
            interval_type = self._config.get_option("intervalType")
        if filters is None:
            filters = self._config.get_option("filters")
        if model_version is None:
            model_version = self._config.get_option("modelVersion")

        predicates = convert_config_filters_to_predicates(filters or [])
        data = self._client.get_feature_percentiles(
            model_id=self.model_id,
            feature_name=feature_name,
            percentiles=percentiles,
            start_date=start_date,
            end_date=end_date,
            interval_type=interval_type,
            predicates=predicates,
            model_version=model_version,
        )
        df = feature_percentiles_to_dataframe(data)
        df.attrs.update(
            {
                "method": "get_feature_percentiles",
                "feature_name": feature_name,
                "start_date": start_date,
                "end_date": end_date,
                "interval_type": interval_type,
                "filters": filters,
                "model_version": model_version,
            }
        )
        return df

    def get_numerical_features(self) -> List[str]:
        """
        Get the list of numerical features available for this model.

        Returns:
            List[str]: List of numerical feature names
        """
        return self._client.get_numerical_features(self.model_id)

    def get_categorical_features(self) -> List[str]:
        """
        Get the list of categorical features available for this model.

        Returns:
            List[str]: List of categorical feature names
        """
        return self._client.get_categorical_features(self.model_id)



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2021-2025 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

from typing import Any, Dict, List, Optional
from urllib.parse import urljoin

import requests

from . import prefix_helpers


class Client:
    """
    A REST client for interacting with the Comet MPM API.

    This client provides methods for making HTTP requests to the Comet MPM API endpoints,
    including model predictions, feature analysis, and custom metric queries.

    Note:
        The client automatically handles authentication for all requests.
    """

    def __init__(
        self,
        request_session: requests.Session,
        base_url: str,
    ) -> None:
        """
        Initialize the Comet MPM REST client.

        Args:
            request_session: The requests session to use for HTTP requests
            base_url: The base URL for the Comet MPM API

        Note:
            The client uses the provided session for all HTTP requests and automatically
            includes authentication headers.
        """
        self.base_url = base_url
        self.session = request_session

    def get(
        self, endpoint: str, params: Optional[Dict[str, Any]] = None
    ) -> Dict[str, Any]:
        """
        Make a GET request to the Comet MPM API.

        Args:
            endpoint: The API endpoint to request (e.g., 'api/mpm/v2/model/numberOfPredictions')
            params: Optional query parameters for the request

        Returns:
            Dict: JSON response from the API

        Raises:
            requests.exceptions.RequestException: If the request fails
            requests.exceptions.HTTPError: If the server returns a non-200 status code
        """
        url = urljoin(self.base_url, endpoint)
        response = self.session.get(url, params=params)
        response.raise_for_status()
        return_data: Dict[str, Any] = response.json()
        return return_data

    def post(
        self, endpoint: str, params: Optional[Dict[str, Any]] = None
    ) -> Dict[str, Any]:
        """
        Make a POST request to the Comet MPM API.

        Args:
            endpoint: The API endpoint to request (e.g., 'api/mpm/v2/features/drift')
            params: Request body parameters as a dictionary

        Returns:
            Dict: JSON response from the API

        Raises:
            requests.exceptions.RequestException: If the request fails
            requests.exceptions.HTTPError: If the server returns a non-200 status code
        """
        url = urljoin(self.base_url, endpoint)
        response = self.session.post(url, json=params)
        response.raise_for_status()
        return_data: Dict[str, Any] = response.json()
        return return_data

    def get_model_id(self, workspace_name: str, model_name: str) -> Optional[str]:
        """
        Retrieve the ID of a specific model within a workspace.

        Args:
            workspace_name: The name of the workspace containing the model
            model_name: The name of the model to search for

        Returns:
            Optional[str]: The ID of the model if found, otherwise None

        Raises:
            requests.exceptions.RequestException: If the request to the API fails
            requests.exceptions.HTTPError: If the server returns a non-200 status code
        """
        endpoint = "api/mpm/v2/model/search"
        params = {
            "page": 1,
            "pageSize": 1000,
            "predicates": [],
            "workspaceId": workspace_name + "-default",  # FIXME after BE update
        }
        response = self.post(endpoint, params)
        model_id: Optional[str] = None
        for model in response["models"]:
            if model["name"] == model_name:
                model_id = model["id"]
        return model_id

    def get_model_details(self, model_id: str) -> Dict[str, Any]:
        """
        Get detailed information about a specific model.

        Args:
            model_id: The ID of the model to retrieve details for

        Returns:
            Dict[str, Any]: Model details including metadata, configuration, and status

        Raises:
            requests.exceptions.HTTPError: If the model is not found or the request fails
        """
        endpoint = f"api/mpm/v2/model/details?modelId={model_id}"
        response = self.get(endpoint)
        return response

    def get_nb_predictions(
        self,
        model_id: str,
        start_date: str,
        end_date: str,
        interval_type: str,
        predicates: List[Dict[str, str]],
        model_version: Optional[str],
    ) -> Dict[str, Any]:
        """
        Get the number of predictions for a model within a specified time range.

        Args:
            model_id: The ID of the model
            start_date: Start date for the query (ISO format)
            end_date: End date for the query (ISO format)
            interval_type: Time interval type ("DAILY" or "HOURLY")
            predicates: List of filter predicate dictionaries with "key" and "query" fields
            model_version: Model version identifier

        Returns:
            Dict[str, Any]: Number of predictions with time series data

        Raises:
            requests.exceptions.HTTPError: If the query fails or returns invalid results
        """
        endpoint = "api/mpm/v2/model/numberOfPredictions"
        params = {
            "modelId": model_id,
            "from": start_date,
            "to": end_date,
            "intervalType": interval_type,
            "predicates": predicates,
            "version": model_version,
        }
        response = self.post(endpoint, params)
        return response

    def get_custom_metrics(
        self,
        model_id: str,
        sql: str,
        start_date: str,
        end_date: str,
        interval_type: str,
        predicates: List[Dict[str, str]],
        model_version: Optional[str],
    ) -> Dict[str, Any]:
        """
        Get custom metrics using SQL query.

        Args:
            model_id: ID of the model
            sql: SQL query string (e.g., "SELECT count(*) FROM model")
            start_date: Start date for the query (ISO format)
            end_date: End date for the query (ISO format)
            interval_type: Time interval type ("DAILY" or "HOURLY")
            predicates: List of filter predicate dictionaries with "key" and "query" fields
            model_version: Model version identifier

        Returns:
            Dict[str, Any]: Metric results from the SQL query

        Raises:
            requests.exceptions.HTTPError: If the query fails or returns invalid results
        """
        endpoint = "api/mpm/v2/custom-metrics/query"
        params = {
            "modelId": model_id,
            "cometSql": sql,
            "from": start_date,
            "to": end_date,
            "intervalType": interval_type,
            "predicates": predicates,
            "version": model_version,
        }
        response = self.post(endpoint, params)
        return response

    def get_feature_drift(
        self,
        model_id: str,
        feature_name: str,
        algorithm: str,
        start_date: str,
        end_date: str,
        interval_type: str,
        predicates: List[Dict[str, str]],
        model_version: Optional[str],
    ) -> Dict[str, Any]:
        """
        Get feature drift analysis between different data sources.

        Args:
            model_id: ID of the model
            feature_name: Name of the feature to analyze
            algorithm: Drift detection algorithm ("EMD", "PSI", or "KL")
            start_date: Start date for the analysis (ISO format)
            end_date: End date for the analysis (ISO format)
            interval_type: Time interval type ("DAILY" or "HOURLY")
            predicates: List of filter predicate dictionaries with "key" and "query" fields
            model_version: Model version identifier

        Returns:
            Dict[str, Any]: Feature drift analysis results

        Raises:
            requests.exceptions.HTTPError: If the analysis fails or returns invalid results
        """
        endpoint = "api/mpm/v2/features/drift"
        feature_name = prefix_helpers.remove_prefix(feature_name)
        source = self.get_feature_source_type(model_id, feature_name)
        if source is None:
            raise Exception("unable to handle label as feature")
        params = {
            "modelId": model_id,
            "name": feature_name,
            "from": start_date,
            "to": end_date,
            "intervalType": interval_type,
            "algorithmType": algorithm,
            "source": source,
            "predicates": predicates,
            "version": model_version,
        }
        response = self.post(endpoint, params)
        return response

    def get_feature_category_distribution(
        self,
        model_id: str,
        feature_name: str,
        normalize: bool,
        start_date: str,
        end_date: str,
        interval_type: str,
        predicates: List[Dict[str, str]],
        model_version: Optional[str],
    ) -> Dict[str, Any]:
        """
        Get the category distribution of a feature.

        Args:
            model_id: Model ID
            feature_name: Name of the feature to analyze
            normalize: Whether to normalize the distribution (default: False)
            start_date: Start date for the analysis (ISO format)
            end_date: End date for the analysis (ISO format)
            interval_type: Time interval type ("DAILY" or "HOURLY")
            predicates: List of filter predicate dictionaries with "key" and "query" fields
            model_version: Model version identifier

        Returns:
            Dict[str, Any]: Category distribution of the feature

        Raises:
            requests.exceptions.HTTPError: If the analysis fails or returns invalid results

        Note:
            This method is specifically for categorical features.
        """
        categorical_features = self.get_categorical_features(model_id)
        if feature_name not in categorical_features:
            raise ValueError(
                f"{feature_name} is not categorical; should be one of {categorical_features}"
            )

        endpoint = "api/mpm/v2/features/distribution"
        feature_name = prefix_helpers.remove_prefix(feature_name)
        source = self.get_feature_source_type(model_id, feature_name)
        params = {
            "modelId": model_id,
            "normalize": normalize,
            "name": feature_name,
            "from": start_date,
            "to": end_date,
            "intervalType": interval_type,
            "source": source,
            "predicates": predicates,
            "version": model_version,
        }
        response = self.post(endpoint, params)
        return response

    def get_feature_density(
        self,
        model_id: str,
        feature_name: str,
        start_date: str,
        end_date: str,
        predicates: List[Dict[str, str]],
        model_version: Optional[str],
        interval_type: str,
    ) -> Dict[str, Any]:
        """
        Get the probability density function (PDF) of a numerical feature.

        Args:
            model_id: ID of the model
            feature_name: Name of the numerical feature
            start_date: Start date for the analysis (ISO format)
            end_date: End date for the analysis (ISO format)
            predicates: List of filter predicate dictionaries with "key" and "query" fields
            model_version: Model version identifier
            interval_type: Time interval type ("DAILY" or "HOURLY")

        Returns:
            Dict[str, Any]: Probability density function of the feature values

        Raises:
            requests.exceptions.HTTPError: If the analysis fails or returns invalid results

        Note:
            This method is specifically for numerical features and returns a PDF.
        """
        numerical_features = self.get_numerical_features(model_id)
        if feature_name not in numerical_features:
            raise ValueError(
                f"{feature_name} is not numerical; should be one of {numerical_features}"
            )

        endpoint = "api/mpm/v2/features/numerical-distribution-pdf"
        feature_name = prefix_helpers.remove_prefix(feature_name)
        source = self.get_feature_source_type(model_id, feature_name)
        params = {
            "modelId": model_id,
            "name": feature_name,
            "from": start_date,
            "to": end_date,
            "intervalType": interval_type,
            "source": source,
            "predicates": predicates,
            "version": model_version,
        }
        response = self.post(endpoint, params)
        return response

    def get_feature_percentiles(
        self,
        model_id: str,
        feature_name: str,
        percentiles: List[float],
        start_date: str,
        end_date: str,
        predicates: List[Dict[str, str]],
        model_version: Optional[str],
        interval_type: str,
    ) -> Dict[str, Any]:
        """
        Get percentile values for a numerical feature.

        Args:
            model_id: ID of the model
            feature_name: Name of the numerical feature
            percentiles: List of percentile values to calculate (e.g., [0.25, 0.5, 0.75])
            start_date: Start date for the analysis (ISO format)
            end_date: End date for the analysis (ISO format)
            predicates: List of filter predicate dictionaries with "key" and "query" fields
            model_version: Model version identifier
            interval_type: Time interval type ("DAILY" or "HOURLY")

        Returns:
            Dict[str, Any]: Percentile values for the specified feature

        Raises:
            requests.exceptions.HTTPError: If the analysis fails or returns invalid results

        Note:
            This method is specifically for numerical features and returns percentile statistics.
        """
        numerical_features = self.get_numerical_features(model_id)
        if feature_name not in numerical_features:
            raise ValueError(
                f"{feature_name} is not numerical; should be one of {numerical_features}"
            )

        endpoint = "api/mpm/v2/features/distribution"
        feature_name = prefix_helpers.remove_prefix(feature_name)
        source = self.get_feature_source_type(model_id, feature_name)
        params = {
            "modelId": model_id,
            "name": feature_name,
            "from": start_date,
            "to": end_date,
            "intervalType": interval_type,
            "source": source,
            "predicates": predicates,
            "version": model_version,
        }
        response = self.post(endpoint, params)
        return response

    def get_categorical_features(self, model_id: str) -> List[str]:
        """
        Get the list of categorical features for a model.

        Args:
            model_id: The ID of the model

        Returns:
            List[str]: List of categorical feature names with their prefixes
        """
        details = self.get_model_details(model_id)
        return [
            prefix_helpers.source_type_to_prefix(feature.get("source"))
            + "_"
            + feature["name"]
            for feature in details["model"]["features"]
            if feature["type"] == "CATEGORICAL"
        ]

    def get_numerical_features(self, model_id: str) -> List[str]:
        """
        Get the list of numerical features for a model.

        Args:
            model_id: The ID of the model

        Returns:
            List[str]: List of numerical feature names with their prefixes
        """
        details = self.get_model_details(model_id)
        return [
            prefix_helpers.source_type_to_prefix(feature.get("source"))
            + "_"
            + feature["name"]
            for feature in details["model"]["features"]
            if feature["type"] == "NUMERICAL"
        ]

    def get_feature_source_type(
        self, model_id: str, feature_name: str
    ) -> Optional[str]:
        """
        Get the source type of a feature.

        Args:
            model_id: ID of the model
            feature_name: Name of the feature (unprefixed)

        Returns:
            Optional[str]: Source type of the feature, or None if not found
        """
        # Unprefixed feature_name
        details = self.get_model_details(model_id)
        matching_sources = [
            feature.get("source")
            for feature in details["model"]["features"]
            if feature["name"] == feature_name
        ]
        return matching_sources[0] if matching_sources else None



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2021 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

import logging
from typing import Any, Callable, Optional, TypeVar

from pydantic import (
    AliasChoices,
    AnyHttpUrl,
    BeforeValidator,
    Field,
    PositiveInt,
    ValidationError,
)
from pydantic_settings import BaseSettings, SettingsConfigDict
from typing_extensions import Annotated

from .logging_messages import INVALID_SETTING_DEFAULT_FALLBACK

LOGGER = logging.getLogger(__name__)

__all__ = ["AnyHttpUrl", "MPMSettings", "MPMSettingsAPI", "LogSettings"]


def log_level_validator(value: Any) -> int:
    if isinstance(value, int):
        if value not in logging._levelToName.keys():
            raise ValueError("Invalid logging level %r" % value)
        return value
    elif not isinstance(value, str):
        raise ValueError("Invalid logging level name %r" % value)
    else:
        level = logging._nameToLevel.get(value.upper(), None)
        if level is None:
            raise ValueError("Invalid logging level name %r" % value)
        return level


LogLevel = Annotated[int, BeforeValidator(log_level_validator)]


class LogSettings(BaseSettings):
    mpm_logging_console: LogLevel = LogLevel(logging.INFO)
    mpm_logging_propagate: bool = False
    mpm_logging_file: Optional[str] = None
    mpm_logging_file_level: Optional[LogLevel] = None
    mpm_logging_file_overwrite: bool = False

    model_config = SettingsConfigDict(env_prefix="comet_")


class MPMSettings(BaseSettings):
    # Use same names as the comet_ml Python SDK to ease configuration
    api_key: str
    url: Optional[AnyHttpUrl] = Field(
        validation_alias=AliasChoices("url", "comet_url", "comet_url_override"),
        default=None,
    )

    mpm_workspace_name: str
    mpm_model_name: str
    mpm_model_version: str

    mpm_max_batch_size: PositiveInt = 1000
    mpm_max_batch_time: PositiveInt = 60
    mpm_join_timeout: PositiveInt = 30
    mpm_batch_sending_timeout: PositiveInt = 5 * 60  # 5 minutes
    mpm_raise_on_error_during_init: bool = False

    model_config = SettingsConfigDict(env_prefix="comet_")


class MPMSettingsAPI(BaseSettings):
    api_key: str
    url: Optional[AnyHttpUrl]


MODEL = TypeVar("MODEL", bound="BaseSettings")


def get_model(m: Callable[..., MODEL], **kwargs: Any) -> MODEL:
    try:
        return m(**kwargs)
    except ValidationError as e:
        for error in e.errors():
            # This should always be a str when dealing with non-nested fields
            error_loc = error["loc"][0]
            assert isinstance(error_loc, str)
            # TODO: Display better error
            # Use the default value instead

            # This is either an extra field or something else that we cannot handle so re-raise original exception
            if error_loc not in m.__fields__:  # type: ignore
                raise

            # The __fields__ attribute is not officially documented
            kwargs[error_loc] = m.__fields__[error_loc].get_default()  # type: ignore

        LOGGER.warning(
            INVALID_SETTING_DEFAULT_FALLBACK,
            exc_info=True,
        )

        return m(**kwargs)



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2021-2025 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************
import logging
import os
import platform
import time
import traceback
from typing import TYPE_CHECKING, Any, Optional, Union

if TYPE_CHECKING:
    from ._logging import ErrorStore

import aiohttp
import requests
from requests import Session
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from . import authhook
from .authhook import aws_sagemaker
from .connection_helpers import is_alive_raise_for_status_code, make_request_headers
from .dataset_upload_helpers import get_request_data
from .logging_messages import BATCH_SENDING_ERROR, UPLOAD_CSV_DATASET_ERROR
from .utils import get_comet_version

LOGGER = logging.getLogger(__name__)

MPM_BASE_PATH = os.environ.get("COMET_MPM_BASE_PATH", "/mpm/")

API_KEY_HEADER = "Authorization"


def setup_http_session_authentication(session: Session) -> None:
    """
    Sets up HTTP session authentication for accessing APIs.

    Args:
        session (Session): The HTTP session object that requires authentication setup.

    This function ensures authentication for accessing AWS SageMaker and integrates authentication hooks into the session.
    It checks if the environment is within AWS SageMaker and performs login operations specific to AWS SageMaker if True.

    Steps:
        1. Checks and performs late login to AWS SageMaker to read environment configuration values that might be set post-import.
        2. Integrates authentication hooks into the provided HTTP session.
    """
    if aws_sagemaker.is_in_aws_sagemaker():
        aws_sagemaker.login_aws_sagemaker()

    authhook.http_session_hook(session=session)


def get_retry_strategy() -> Retry:
    # The total backoff sleeping time is computed like that:
    # backoff = 2
    # total = 3
    # s = lambda b, i: b * (2 ** (i - 1))
    # sleep = sum(s(backoff, i) for i in range(1, total + 1))

    return Retry(total=3, backoff_factor=2, allowed_methods=None)  # Will wait up to 24s


def get_comet_http_session(
    api_key: Optional[str], retry_strategy: Optional[Retry] = None
) -> Session:
    # TODO: Check if config is needed
    # """Creates http session with Comet related headers set and authentication hook enabled."""
    # if config is None:
    #     config = get_config()  # This can be slow if called for every new session

    session = get_http_session(
        retry_strategy=retry_strategy,
        # config=config,
        # TODO: Check if verify_tls is needed
        # TODO: Check if tcp_keep_alive is needed
    )

    session.headers.update(
        {
            "X-COMET-DEBUG-SDK-VERSION": get_comet_version(),
            "X-COMET-DEBUG-PY-VERSION": platform.python_version(),
        }
    )

    # Add authorization header
    if api_key is not None:
        session.headers.update({API_KEY_HEADER: api_key})

    # Setup HTTP allow header if configured
    # TODO: Check if config is needed
    # allow_header_name = config["comet.allow_header.name"]
    # allow_header_value = config["comet.allow_header.value"]

    # if allow_header_name and allow_header_value:
    #     session.headers[allow_header_name] = allow_header_value

    setup_http_session_authentication(session)

    return session


def get_api_http_session(api_key: str) -> requests.Session:
    """
    Get a Request Session for MPM SDK API.
    """
    # The total backoff sleeping time is computed like that:
    # backoff = 2
    # total = 3
    # s = lambda b, i: b * (2 ** (i - 1))
    # sleep = sum(s(backoff, i) for i in range(1, total + 1))

    status_codes = [429, 500, 502, 503, 504]
    retry_strategy = Retry(
        total=3,
        status_forcelist=status_codes,
        backoff_factor=2,
        raise_on_status=False,
    )
    session = get_http_session(retry_strategy)
    # Add authorization header
    if api_key is not None:
        session.headers.update({API_KEY_HEADER: api_key})
    return session


def get_http_session(retry_strategy: Optional[Retry] = None) -> Session:
    session = Session()

    # Setup retry strategy if asked
    if retry_strategy is not None:
        session.mount("http://", HTTPAdapter(max_retries=retry_strategy))
        session.mount("https://", HTTPAdapter(max_retries=retry_strategy))

    return session


def stream_dataset_file(
    api_key: str,
    file_path: str,
    base_url: Optional[str],
    workspace_name: str,
    model_name: str,
    model_version: str,
    dataset_type: str,
    dataset_name: Optional[str] = None,
    na_values: Optional[str] = None,
    keep_default_na_values: Optional[str] = None,
) -> None:
    """Streams a file to a server with additional parameters and prints the response progress_message
    by progress_message using a requests.Session."""
    start_time = time.time()  # Start timing
    response = None
    try:
        # Open the file in binary mode for streaming
        with open(file_path, "rb") as file:
            files = {"file": (file_path, file, "text/csv")}
            data, url = get_request_data(
                base_url,
                workspace_name,
                model_name,
                model_version,
                dataset_type,
                dataset_name,
                na_values,
                keep_default_na_values,
            )

            headers = {"Authorization": api_key}
            # Stream the file to the server with additional data
            response = requests.post(
                headers=headers,
                url=url,
                files=files,
                data=data,
                stream=True,
            )

            response.raise_for_status()
            for progress_message in response.iter_lines():
                LOGGER.info(progress_message.decode("utf-8"))

            end_time = time.time()  # End timing
            duration = end_time - start_time
            LOGGER.info(f"Finished upload in: {duration:.2f} seconds")
    except requests.exceptions.RequestException as e:
        if response is None:
            LOGGER.error(f"{UPLOAD_CSV_DATASET_ERROR}. Error: {str(e)}")
        else:
            LOGGER.error(
                f"{UPLOAD_CSV_DATASET_ERROR}. Error: {response.status_code}, {response.text}. {str(e)}",
                exc_info=True,
            )
    except Exception as e:
        LOGGER.error(f"{UPLOAD_CSV_DATASET_ERROR}. Error: {str(e)}")


def send_batch_requests(
    session: Session,
    url_endpoint: str,
    api_key: str,
    batch_sending_timeout: int,
    batch: Any,
    error_store: Optional["ErrorStore"] = None,
) -> None:
    response: Union[requests.Response, None] = None
    try:
        headers = make_request_headers(api_key)
        json = {"data": batch}
        response = session.post(
            url_endpoint, headers=headers, json=json, timeout=batch_sending_timeout
        )
        response.raise_for_status()
        # TODO: Process response
    except requests.exceptions.RequestException as e:
        if response is not None:
            error_message = (
                f"{BATCH_SENDING_ERROR}. Error: {response.status_code}, {response.text}"
            )
            LOGGER.error(error_message)

            if error_store is not None:
                error_store.add_error(
                    message=error_message,
                    logger_name="comet_mpm.connection",
                    data_affected=batch,
                    traceback_info=traceback.format_exc(),
                )
        else:
            error_message = f"{BATCH_SENDING_ERROR}. Error: {str(e)}"
            LOGGER.error(error_message)
            if error_store is not None:
                error_store.add_error(
                    message=error_message,
                    logger_name="comet_mpm.connection",
                    data_affected=batch,
                    traceback_info=traceback.format_exc(),
                )

    except Exception as e:
        error_message = f"{BATCH_SENDING_ERROR}. Error: {str(e)}"
        LOGGER.error(error_message)

        if error_store is not None:
            error_store.add_error(
                message=error_message,
                logger_name="comet_mpm.connection",
                data_affected=batch,
                traceback_info=traceback.format_exc(),
            )


def send_is_alive(
    session: Session,
    url_endpoint: str,
    api_key: str,
) -> None:
    """Raise CometMPMBackendException if 500 status code received.
    Raise CometMPMBackendIsNotAvailable if any other error code received.
    According to: https://comet-ml.atlassian.net/browse/CM-9503?focusedCommentId=32968"""
    headers = make_request_headers(api_key)
    response = session.get(url_endpoint, headers=headers)
    is_alive_raise_for_status_code(status_code=response.status_code)


async def send_asyncio_batch_requests(
    session: aiohttp.ClientSession,
    url_endpoint: str,
    api_key: str,
    batch_sending_timeout: int,
    batch: Any,
    error_store: Optional["ErrorStore"] = None,
) -> None:
    try:
        headers = make_request_headers(api_key)
        json = {"data": batch}
        timeout = aiohttp.ClientTimeout(total=batch_sending_timeout)
        response = await session.post(
            url_endpoint, headers=headers, json=json, timeout=timeout
        )
        response.raise_for_status()
    except Exception as e:
        error_message = f"{BATCH_SENDING_ERROR}. Error: {str(e)}"
        LOGGER.error(error_message)

        if error_store is not None:
            error_store.add_error(
                message=error_message,
                logger_name="comet_mpm.connection",
                data_affected=batch,
                traceback_info=traceback.format_exc(),
            )


async def send_asyncio_is_alive(
    session: aiohttp.ClientSession,
    url_endpoint: str,
    api_key: str,
) -> None:
    headers = make_request_headers(api_key)
    response = await session.get(url_endpoint, headers=headers)
    is_alive_raise_for_status_code(status_code=response.status)



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.com
#  Copyright (C) 2015-2025 Comet ML INC
#  This source code is licensed under the MIT license.
# *******************************************************
import logging
import os
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from sagemaker.partner_app.auth_provider import PartnerAppAuthProvider

from requests import Session

from comet_mpm import authhook

LOGGER = logging.getLogger(__name__)

AWS_SAGEMAKER_AUTH_INITIALIZED = False


def is_in_aws_sagemaker() -> bool:
    return os.getenv("AWS_PARTNER_APP_AUTH") is not None


# TODO: Fix the type hinting, currently it causes
#  src/comet_mpm/authhook/aws_sagemaker.py:62: error: Incompatible types in assignment (expression has type "Callable[[Session], None]", variable has type "Callable[[Arg(Session, 'session')], None]")  [assignment]
def _aws_sagemaker_session_hook(auth_provider: "PartnerAppAuthProvider") -> Any:
    # def _aws_sagemaker_session_hook(auth_provider: "PartnerAppAuthProvider") -> Callable[[Session], None]:

    def _aws_session_hook(session: Session) -> None:
        session.auth = auth_provider.get_auth()

    return _aws_session_hook


def login_aws_sagemaker() -> None:
    """
    Initializes AWS SageMaker authentication.

    If the system is not in AWS SageMaker or the authentication is already set, the function will
    return immediately without performing any action.

    Sets up SageMaker Partner SDK authentication provider, hooks HTTP session with the AWS SageMaker authentication
    provider, and marks the authentication as initialized.
    """
    global AWS_SAGEMAKER_AUTH_INITIALIZED

    if not is_in_aws_sagemaker() or AWS_SAGEMAKER_AUTH_INITIALIZED:
        return

    LOGGER.debug("AWS partner SDK authentication initialization commenced")

    # setup sagemaker partner SDK authentication provider
    from sagemaker.partner_app.auth_provider import PartnerAppAuthProvider

    auth_provider = PartnerAppAuthProvider()
    authhook.http_session_hook = _aws_sagemaker_session_hook(
        auth_provider=auth_provider
    )

    AWS_SAGEMAKER_AUTH_INITIALIZED = True

    LOGGER.debug("AWS partner SDK authentication initialized")


def logout_aws_sagemaker() -> None:
    global AWS_SAGEMAKER_AUTH_INITIALIZED
    if AWS_SAGEMAKER_AUTH_INITIALIZED:
        authhook.http_session_hook = lambda session: None
        AWS_SAGEMAKER_AUTH_INITIALIZED = False
        LOGGER.debug("AWS partner SDK authentication removed")



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2021-2024 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************
from typing import Dict
from urllib.parse import urljoin, urlparse, urlsplit, urlunparse, urlunsplit

from comet_mpm.exceptions import CometMPMBackendException, CometMPMBackendIsNotAvailable
from comet_mpm.logging_messages import BACKEND_500_ERROR


def sanitize_url(url: str) -> str:
    """Sanitize a URL, checking that it is a valid URL and ensure it contains an ending slash /"""
    parts = urlparse(url)
    scheme, netloc, path, params, query, fragment = parts

    # TODO: Raise an exception if params, query and fragment are not empty?

    # Ensure the leading slash
    if path and not path.endswith("/"):
        path = path + "/"
    elif not path and not netloc.endswith("/"):
        netloc = netloc + "/"

    return urlunparse((scheme, netloc, path, params, query, fragment))


def url_join(base: str, *parts: str) -> str:
    """Given a base and url parts (for example [workspace, project, id]) returns a full URL"""
    # TODO: Enforce base to have a scheme and netloc?
    result = base

    for part in parts[:-1]:
        if not part.endswith("/"):
            raise ValueError("Intermediary part not ending with /")

        result = urljoin(result, part)

    result = urljoin(result, parts[-1])

    return result


def get_root_url(url: str) -> str:
    """Remove the path, params, query and fragment from a given URL"""
    parts = urlsplit(url)
    scheme, netloc, path, query, fragment = parts

    return urlunsplit((scheme, netloc, "", "", ""))


def make_request_headers(api_key: str) -> Dict[str, str]:
    return {"Authorization": api_key}


def is_alive_raise_for_status_code(status_code: int) -> None:
    if status_code == 200:
        return
    if status_code == 500:
        raise CometMPMBackendException(BACKEND_500_ERROR)
    else:
        raise CometMPMBackendIsNotAvailable()



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2025 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

import json
from collections import defaultdict
from typing import Any, Dict, List, Tuple

import pandas as pd


def custom_metrics_to_dataframe(data: Dict[str, Any], sql: str) -> pd.DataFrame:
    """
    Convert the given dict into a pandas DataFrame.
    """
    # Collect the dates
    dates = [point["x"] for point in data["data"][0]["data"]]
    df = pd.DataFrame({"timestamp": pd.to_datetime(dates)})

    # For each series, add a column
    sql_set = set()
    series_data = {}
    for series in data["data"]:
        if "data" not in series:
            continue
        # FIXME: only allow one select statement, and get it from parameter
        colname = sql
        sql_set.add(colname)
        predkey = series.get("predicateKey", "predicate")
        y_values = [point["y"] for point in series["data"]]
        key = (predkey, colname)
        series_data[key] = y_values

    series_df = pd.DataFrame(series_data)
    df = pd.concat([df, series_df], axis=1)
    df.set_index("timestamp", inplace=True)
    if len(sql_set) == 1:
        column_names: List[str] = [
            str(name).replace("'", "")
            if isinstance(name, str)
            else str(name[0]).replace("'", "")
            for name in df.columns
        ]
        df.columns = column_names
    else:
        column_tuples: List[Tuple[str, str]] = [
            (str(name[0]).replace("'", ""), str(name[1]).replace("'", ""))
            for name in df.columns
        ]
        df.columns = pd.MultiIndex.from_tuples(
            column_tuples, names=["predicateKey", "columnName"]
        )
    return df


def feature_drift_to_dataframe(data: Dict[str, Any]) -> pd.DataFrame:
    rows: Dict[Any, Dict[str, Any]] = defaultdict(dict)
    for filter_dict in data["data"]:
        filter_name = filter_dict["predicateKey"]
        for xy in filter_dict["data"]:
            key = xy["x"]  # timestamp
            rows[key][filter_name] = xy["y"]
    # Unwrap
    new_rows = []
    for key in rows:
        row = {"timestamp": key}
        for filter_name in rows[key]:
            row[filter_name] = rows[key][filter_name]
        new_rows.append(row)
    df = pd.DataFrame(new_rows)
    if "timestamp" in df.columns:
        df = df.set_index("timestamp")
    return df


def feature_category_distribution_to_dataframe(data: Dict[str, Any]) -> pd.DataFrame:
    # First, collect the filter columns:
    rows: Dict[str, Dict[str, Any]] = defaultdict(dict)
    for filter_dict in data["data"]:
        for chart_data in filter_dict["chartData"]:
            for point in chart_data["points"]:
                key = json.dumps((chart_data["value"], point["x"]))
                rows[key].update({filter_dict["predicateKey"]: point["y"]})
    # Next, unwrap into rows:
    new_rows = []
    for key in rows:
        value, x = json.loads(key)
        new_row = {
            "timestamp": x,
            "value": value,
        }
        new_row.update(rows[key])
        new_rows.append(new_row)
    df = pd.DataFrame(new_rows)
    if "timestamp" in df.columns:
        df = df.set_index("timestamp")
    return df


def feature_density_to_dataframe(data: Dict[str, Any]) -> pd.DataFrame:
    rows: Dict[Any, Dict[str, Any]] = defaultdict(dict)
    for item in data["data"]:
        predicate = item["predicateKey"]
        for point in item["pdfDistributionGraphs"]["wholeTimeRangeDistribution"][
            "graphPoints"
        ]:
            key = point["x"]
            rows[key][predicate] = point["y"]
    new_rows = []
    for key in rows:
        row = {"x": key}
        for item in rows[key]:
            row[item] = rows[key][item]
        new_rows.append(row)
    df = pd.DataFrame(new_rows)
    if "x" in df.columns:
        df = df.set_index("x")
        df.sort_values(by="x", inplace=True)

    return df


def feature_percentiles_to_dataframe(data: Dict[str, Any]) -> pd.DataFrame:
    rows: Dict[Tuple[str, str], Dict[str, Any]] = defaultdict(dict)
    for item in data["data"]:
        for chart_data in item["chartData"]:
            predicate_key = item["predicateKey"]
            for point in chart_data["points"]:
                timestamp = point["x"]
                key = (timestamp, chart_data["value"])
                rows[key][predicate_key] = point["y"]
    new_rows = []
    for key in rows:
        timestamp, percentile = key
        row = {
            "timestamp": timestamp,
            "percentile": percentile,
        }
        for predicate in rows[key]:
            row[predicate] = rows[key][predicate]
        new_rows.append(row)
    df = pd.DataFrame(new_rows)
    if "timestamp" in df.columns:
        df = df.set_index("timestamp")
    return df



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2021 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

import asyncio as asyncio_module
import atexit
import logging
import os
import sys
import traceback
from typing import Any, Awaitable, Dict, Iterable, List, Optional, Union

from . import constants, logging_messages, optional_update
from ._logging import ErrorStore
from .connection import MPM_BASE_PATH, stream_dataset_file
from .connection_helpers import url_join
from .environment import check_environment
from .events import events_from_dataframe
from .events.label_event import LabelsEvent
from .events.prediction_event import PredictionEvent
from .exceptions import CometMPMBackendException, CometMPMBackendIsNotAvailable
from .logging_messages import MPM_JOIN_DEPRECATED_WARNING, MPM_LABEL_DEPRECATED_WARNING
from .sender import get_sender
from .server_address import ServerAddress
from .settings import MPMSettings, get_model
from .settings_helper import extract_comet_url

LOGGER = logging.getLogger(__name__)

LogEventsResult = Union[List[Any], Awaitable[List[Any]]]


class CometMPM:
    """
    The Comet MPM class is used to upload a model's input and output features to MPM
    """

    def __init__(
        self,
        api_key: Optional[str] = None,
        workspace_name: Optional[str] = None,
        model_name: Optional[str] = None,
        model_version: Optional[str] = None,
        disabled: Optional[bool] = None,
        asyncio: bool = False,
        max_batch_size: Optional[int] = None,
        max_batch_time: Optional[int] = None,
        raise_on_error_during_init: bool = False,
    ):
        """
        Creates the Comet MPM Event logger object.
        Args:
            api_key: The Comet API Key
            workspace_name: The Comet Workspace Name of the model
            model_name: The Comet Model Name of the model
            model_version: The Comet Model Version of the model
            disabled: If set to True, CometMPM will not send anything to the backend.
            asyncio: Set to True if you are using an Asyncio-based framework like FastAPI.
            max_batch_size: Maximum number of MPM events sent in a batch, can also be configured using the environment variable MPM_MAX_BATCH_SIZE.
            max_batch_time: Maximum time before a batch of events is submitted to MPM, can also be configured using the environment variable MPM_MAX_BATCH_SIZE.
            raise_on_error_during_init: If set to True, CometMPM will raise exceptions instead of just logging errors during initialization. Default is False for backwards compatibility.
        """

        settings_user_values: Dict[str, Union[str, int]] = {}
        optional_update.update(
            settings_user_values,
            {
                "api_key": api_key,
                "mpm_model_name": model_name,
                "mpm_model_version": model_version,
                "mpm_workspace_name": workspace_name,
                "mpm_max_batch_size": max_batch_size,
                "mpm_max_batch_time": max_batch_time,
            },
        )

        self._settings = get_model(
            MPMSettings,
            **settings_user_values,
        )
        if disabled:
            self.disabled = disabled  # type: bool
        else:
            self.disabled = bool(os.getenv("COMET_MPM_DISABLED"))
        self._asyncio = asyncio
        # Use the provided raise_on_error_during_init value, or fall back to settings value
        self._raise_on_error_during_init = (
            raise_on_error_during_init or self._settings.mpm_raise_on_error_during_init
        )

        comet_url = extract_comet_url(self._settings)

        self._mpm_server_address = ServerAddress(
            base_url=url_join(comet_url, MPM_BASE_PATH),
            api_version="v2",
        )

        # Create instance-specific error store for capturing background thread errors
        self._error_store = ErrorStore()

        if self.disabled:
            self._sender = None
        else:
            self._sender = get_sender(
                api_key=self._settings.api_key,
                server_address=self._mpm_server_address,
                max_batch_size=self._settings.mpm_max_batch_size,
                max_batch_time=self._settings.mpm_max_batch_time,
                asyncio=self._asyncio,
                batch_sending_timeout=self._settings.mpm_batch_sending_timeout,
                error_store=self._error_store,
            )

            if not self._asyncio:
                self._do_connect()

            atexit.register(self._on_end)

        check_environment()

    def log_event(
        self,
        prediction_id: str,
        input_features: Optional[Dict[str, Any]] = None,
        output_value: Optional[Any] = None,
        output_probability: Optional[Any] = None,
        output_features: Optional[Dict[str, Any]] = None,
        labels: Optional[Dict[str, Union[int, float, bool, str]]] = None,
        timestamp: Optional[float] = None,
    ) -> Optional[Awaitable[None]]:
        """
        Asynchronously log a single event to MPM. Events are identified by the
        mandatory prediction_id parameter.
        If you send multiple events with the same prediction_id,
        the Comet platform with automatically reject the duplicate events.

        Args:
            prediction_id: The unique prediction ID. It can be provided by the
                framework, you, or a random unique value such as str(uuid4()).
            input_features: If provided, it must be a flat dictionary where the
                keys are the feature names, and the values are native Python
                scalars, such as integers, floats, booleans, or strings. For
                example: `{"age": 42, "income": 42894.89}`.
            output_value: The prediction as a native Python scalar, such as an
                 integer, float, boolean, or string.
            output_probability: If provided, it must be a float between 0 and
                 1, indicating the model's confidence in the prediction.
            output_features: A dictionary of output features.
            labels: If provided, it must be a flat dictionary where the
                keys are the label names, and the values are native Python
                scalars, such as integers, floats, booleans, or strings. For
                example: `{"person": 2, "bicycle": 1, "car": 3}`.
            timestamp: An optional timestamp to associate with the event
                (seconds since epoch in UTC timezone). If not provided, the
                 current time will be used."""
        if self.disabled:
            if self._asyncio is False:
                return None
            else:
                return asyncio_module.sleep(0)

        output_features = _handle_event_output_features(
            output_value, output_probability, output_features
        )

        event = PredictionEvent(
            workspace=self._settings.mpm_workspace_name,
            model_name=self._settings.mpm_model_name,
            model_version=self._settings.mpm_model_version,
            prediction_id=prediction_id,
            input_features=input_features,
            output_features=output_features,
            labels=labels,
            timestamp=timestamp,
        )
        return self._log_event(event)

    def upload_dataset_csv(
        self,
        file_path: str,
        dataset_type: str,
        dataset_name: Optional[str] = None,
        na_values: Optional[str] = None,
        keep_default_na_values: Optional[str] = None,
    ) -> None:
        """Uploads dataset from a local CSV file to the backend, streaming the data line by line.

        This method facilitates the creation or updating of a dataset with data streamed from a specified
        CSV file. The uploaded data is used as a reference for detecting drift in a production model within the same
        workspace. Each line of the CSV file is sent as an event to the dataset, allowing for incremental
        updates.

        Args:
            file_path (`str`): The path to the local CSV file whose data is to be streamed to the backend.
            dataset_type (`Literal['EVENTS', 'LATE_LABELS', 'TRAINING_EVENTS']`): Type of the dataset to be updated
            or created. Default is 'TRAINING_EVENTS'.
            dataset_name (`str`): The name of the dataset where the data will be stored. If a model
                with this name does not exist, a new model will be created. If the model already exists, new records
                will be added to it (duplicated predictionIds will be ignored).
                In case dataset_type is TRAINING_EVENTS this is mandatory as the MPM model_name
                is the production model, which the dataset_name parameter is referring to.
            na_values (`str`, *optional*): Additional strings to recognize as NA/NaN. By default, the system
                recognizes standard missing values (like empty fields, 'NaN', 'NULL', etc.). Specifying this parameter
                allows for the inclusion of custom missing value identifiers, enhancing the flexibility in data
                handling. If specified, it should be as comma delimiter string. The default list is (Note that empty
                string is also in this list): None,,null,NULL,N/A,NA,NaN,n/a,nan
            keep_default_na_values (`str`, *optional*, defaults to None): A boolean that determines whether to
                include the default set of NA identifiers in addition to the values specified in 'na_values'. If
                `True`, both default and specified missing value identifiers are used. If `False`, only the values
                specified in 'na_values' are considered.

        Returns: None

        Notes:
            CSV Format:
            - The first line of the CSV file must contain headers.
            - Columns:
              1. timestamp (optional): If missing, the current timestamp will be used as the event time. If specified,
                 it should be the millis since epoch.
              2. predictionId (optional): Unique identifier for each event. If missing, a UUID will be generated.
                 Duplicate predictionIds in new events will be ignored.
              3. feature_* columns: These prefixed columns specify the input features for the model,
                 e.g., 'feature_age' or 'feature_color'.
              4. prediction_* columns: These prefixed columns are for the output features,
                 e.g., 'prediction_animal' or 'prediction_probability'.
              5. label_value_* columns: These columns are for the label values of the event,
                 e.g., 'label_value_price' or 'label_value_animal'.

            Sample CSV content:
                timestamp,predictionId,feature_oneMoreFeature,feature_anotherFeature,feature_someFeature,prediction_fingers_count,prediction_probability,prediction_value,label_value_fingers_count,label_value_animal
                1713006000001,someAssetId_-1895825684,Dog,special,53.09863247819340,7,0.87,Bird,4,Fish
                1713006600001,someAssetId_926457604,null,special,55.73110218323990,1,0.69,Fish,6,Fish
                1713007200001,someAssetId_2145792990,Rabbit,special,49.40627545548700,4,0.59,Bird,1,Fish

        Examples:
            ```python linenums="1"
            from comet_mpm import CometMPM

            MPM = CometMPM()
            MPM.upload_dataset_csv(
                file_path="path/to/your/data.csv",
                dataset_type="TRAINING_EVENTS",  # Or use 'EVENTS', 'LATE_LABELS' as needed
                dataset_name="your-dataset-name"
            )
            ```
        """
        stream_dataset_file(
            api_key=self._settings.api_key,
            file_path=file_path,
            base_url=self._mpm_server_address.base_url,
            workspace_name=self._settings.mpm_workspace_name,
            model_name=self._settings.mpm_model_name,
            model_version=self._settings.mpm_model_version,
            dataset_type=dataset_type,
            dataset_name=dataset_name,
            na_values=na_values,
            keep_default_na_values=keep_default_na_values,
        )

    def log_label(
        self,
        prediction_id: str,
        label: Optional[Any] = None,
        labels: Optional[Dict[str, Union[int, float, bool, str]]] = None,
        timestamp: Optional[float] = None,
    ) -> Optional[Awaitable[None]]:
        """
        Send an MPM event containing the ground truth value for a prediction whose input and output
        features are already stored in Comet.
        If you send multiple labels with the same prediction_id,
        the Comet platform with automatically reject the duplicate labels.
        Args:
            prediction_id: The unique prediction ID
            label: Deprecated, please use the labels instead. If provided, this
                value will be used put as 'value' within the labels.
            labels: The ground truth values for the prediction. It must be a flat dictionary where the
                keys are the label names, and the values are native Python
                scalars, such as integers, floats, booleans, or strings. For
                example: `{"person": 2, "bicycle": 1, "car": 3}`.
            timestamp: An optional timestamp to associate with the label
                (seconds since epoch in UTC timezone). If not provided, the
                 current time will be used.
        """
        if self.disabled:
            if self._asyncio is False:
                return None
            else:
                return asyncio_module.sleep(0)

        if labels is None:
            labels = {}

        if label is not None:
            LOGGER.warning(MPM_LABEL_DEPRECATED_WARNING)
            labels["value"] = label

        event = LabelsEvent(
            workspace=self._settings.mpm_workspace_name,
            model_name=self._settings.mpm_model_name,
            model_version=self._settings.mpm_model_version,
            prediction_id=prediction_id,
            labels=labels,
            timestamp=timestamp,
        )
        return self._log_event(event)

    def log_dataframe(  # type: ignore[no-untyped-def]
        self,
        dataframe,
        prediction_id_column: str,
        feature_columns: Optional[List[str]] = None,
        output_value_column: Optional[str] = None,
        output_probability_column: Optional[str] = None,
        output_features_columns: Optional[List[str]] = None,
        labels_columns: Optional[List[str]] = None,
        timestamp_column: Optional[str] = None,
    ) -> LogEventsResult:
        """
        This function logs each row of a Pandas DataFrame as an MPM event. The
        events are structured as described in the [log_event](#cometmpmlog_event)
        method, so please refer to it for full context.

        Args:
            dataframe: The Pandas DataFrame to be logged.
            prediction_id_column: This column should contain the prediction_id values for the
                events.
            feature_columns: If provided, these columns will be used as the input_features
                for the events.
            output_features_columns: If provided, these columns will be used as the output_features for the events.
            output_value_column: Deprecated, please use the output_features_column field instead. If provided, this
                column will be used as the output_value for the events.
            output_probability_column: Deprecated, please use the output_features_column field instead.
                If provided, this column will be used as the output_probability for the events.
            labels_columns: If provided, these columns will be used as the labels for the events.
            timestamp_column: If provided, this column will be used as the timestamp (seconds since
                epoch start in UTC timezone) for the events.
        """
        events = events_from_dataframe.generate(
            workspace=self._settings.mpm_workspace_name,
            model_name=self._settings.mpm_model_name,
            model_version=self._settings.mpm_model_version,
            dataframe=dataframe,
            prediction_id_column=prediction_id_column,
            feature_columns=feature_columns,
            output_features_columns=output_features_columns,
            output_value_column=output_value_column,
            output_probability_column=output_probability_column,
            labels_columns=labels_columns,
            timestamp_column=timestamp_column,
        )

        return self._log_events(events)

    def connect(self) -> Optional[Awaitable[None]]:
        """
        When using CometMPM in asyncio mode, this coroutine needs to be awaited
        at the server start.
        """
        if self._asyncio is False:
            return None
        else:
            if self._sender is not None:
                return asyncio_module.create_task(self._do_async_connect())

            return asyncio_module.sleep(0)

    def join(self, timeout: Optional[int] = None) -> Optional[Awaitable[None]]:
        """
        MPM.join is deprecated, use MPM.end instead.
        """
        LOGGER.warning(MPM_JOIN_DEPRECATED_WARNING)
        return self.end(timeout)

    def end(self, timeout: Optional[int] = None) -> Optional[Awaitable[None]]:
        """Ensure that all data has been sent to Comet and close the MPM object.
        After that, no data can be logged anymore. Waits for up to 30 seconds if timeout is not set.
        """
        if timeout is None:
            timeout = self._settings.mpm_join_timeout

        if not self.disabled:
            assert self._sender is not None
            if self._asyncio:
                return self._sender.join(timeout)
            else:
                self._sender.close(timeout)

                # Check for remaining events after manual close (similar to _on_end)
                if self._sender.sender_queue is not None:
                    remaining_events = self._sender.sender_queue.qsize()
                    if remaining_events > 0:
                        error_message = f"Upload timeout reached, {remaining_events:,.0f} MPM events were not synced with MPM. Consider increasing timeout parameter or using MPM.end() earlier in your script."
                        LOGGER.warning(error_message)
                        print(
                            f"COMET WARNING: {error_message}",
                            file=sys.stderr,
                            flush=True,
                        )

        if self._asyncio is False:
            return None
        else:
            return asyncio_module.sleep(0)

    def get_logging_errors(self, clear: bool = True) -> List[Dict[str, Any]]:
        """Get any logging errors that occurred during background processing.

        This method allows users to programmatically check for and handle
        errors that occurred in background threads, such as network failures
        when sending batch data to the backend.

        Args:
            clear: If True, clear the error store after retrieving errors.
                   Defaults to True to prevent memory leaks.

        Returns:
            List of error dictionaries, each containing:
            - 'message': The error message
            - 'logger_name': Name of the logger that produced the error
            - 'timestamp': ISO timestamp when the error occurred
            - 'data_affected': Description of what data was affected (optional)
            - 'traceback': Exception traceback information (optional)

        Example:
            >>> mpm = CometMPM(...)
            >>> mpm.log_event(...)
            >>> mpm.end()
            >>> errors = mpm.get_logging_errors()
            >>> if errors:
            ...     print(f"Found {len(errors)} errors:")
            ...     for error in errors:
            ...         print(f"  {error['timestamp']}: {error['message']}")
        """
        return self._error_store.get_errors(clear=clear)

    def has_logging_error(self) -> bool:
        """Check if there are any stored logging errors without retrieving them.

        Returns:
            True if there are errors in the store, False otherwise.

        Example:
            >>> mpm = CometMPM(...)
            >>> mpm.log_event(...)
            >>> mpm.end()
            >>> if mpm.has_logging_error():
            ...     errors = mpm.get_logging_errors()
            ...     # Handle errors
        """
        return self._error_store.has_errors()

    def _on_end(self) -> None:
        if not self.disabled:
            assert self._sender is not None
            self._sender.close(timeout=self._settings.mpm_join_timeout)

            if self._sender.sender_queue is not None:
                remaining_events = self._sender.sender_queue.qsize()
                if remaining_events > 0:
                    error_message = f"Upload process timed-out, {remaining_events:,.0f} MPM events were not synced with MPM. Timeout can be increased using the COMET_MPM_JOIN_TIMEOUT environment variable."
                    print(f"COMET ERROR: {error_message}", file=sys.stderr, flush=True)

    def _log_events(self, events: Iterable[PredictionEvent]) -> LogEventsResult:
        results = []
        for event in events:
            result = self._log_event(event)
            if result is not None:
                results.append(result)

        if self._asyncio and len(results) > 0:
            return asyncio_module.gather(*results)

        return results

    def _log_event(
        self, event: Union[PredictionEvent, LabelsEvent]
    ) -> Optional[Awaitable[None]]:
        assert self._sender is not None
        return self._sender.put(event)

    def _do_connect(self) -> None:
        try:
            assert self._sender is not None
            self._sender.connect()
        except (CometMPMBackendException, CometMPMBackendIsNotAvailable) as exc:
            error_message = f"Connection failed during initialization: {str(exc)}"
            LOGGER.error(error_message)
            self.disabled = True

            # Store connection error for programmatic access
            if hasattr(self, "_error_store") and self._error_store is not None:
                self._error_store.add_error(
                    message=error_message,
                    logger_name="comet_mpm.comet_mpm",
                    data_affected=None,
                    traceback_info=traceback.format_exc(),
                )

            if self._raise_on_error_during_init:
                raise

    async def _do_async_connect(self) -> None:
        try:
            assert self._sender is not None
            await self._sender.connect()
        except (CometMPMBackendException, CometMPMBackendIsNotAvailable) as exc:
            LOGGER.error(str(exc))
            self.disabled = True

            # Store connection error for programmatic access
            if hasattr(self, "_error_store") and self._error_store is not None:
                self._error_store.add_error(
                    message=str(exc),
                    logger_name="comet_mpm.comet_mpm",
                    data_affected=None,
                    traceback_info=traceback.format_exc(),
                )

            if self._raise_on_error_during_init:
                raise
        except Exception:
            # Convert connection errors to CometMPMBackendIsNotAvailable for consistency
            backend_exc = CometMPMBackendIsNotAvailable()
            LOGGER.error(str(backend_exc))
            self.disabled = True

            # Store connection error for programmatic access
            if hasattr(self, "_error_store") and self._error_store is not None:
                self._error_store.add_error(
                    message=str(backend_exc),
                    logger_name="comet_mpm.comet_mpm",
                    data_affected=None,
                    traceback_info=traceback.format_exc(),
                )

            if self._raise_on_error_during_init:
                raise backend_exc


def _handle_event_output_features(
    output_value: Any,
    output_probability: Any,
    output_features: Optional[Dict[str, Any]],
) -> Optional[Dict[str, Any]]:
    event_output_features: Optional[Dict[str, Any]]

    if (
        output_value is not None or output_probability is not None
    ) and output_features is None:
        LOGGER.warning(
            logging_messages.DEPRECATED_OUTPUT_VALUE_AND_PROBABILITY_WITHOUT_FEATURES
        )

        event_output_features = {}
        if output_value is not None:
            event_output_features[constants.EVENT_PREDICTION_VALUE] = output_value

        if output_probability is not None:
            event_output_features[
                constants.EVENT_PREDICTION_PROBABILITY
            ] = output_probability
    elif (
        output_value is not None and output_probability is not None
    ) and output_features is not None:
        LOGGER.warning(
            logging_messages.DEPRECATED_OUTPUT_VALUE_AND_PROBABILITY_WITH_FEATURES
        )

        event_output_features = {
            constants.EVENT_PREDICTION_VALUE: output_value,
            constants.EVENT_PREDICTION_PROBABILITY: output_probability,
        }
        for key in output_features:
            event_output_features[key] = output_features[key]
    else:
        event_output_features = output_features

    return event_output_features



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2021 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

from typing import Any, Dict


def update(destination: Dict[str, Any], source: Dict[str, Any]) -> None:
    """
    the equivalent of destination.update(source),
    except that if a (key, value) pair in source is None,
    then this key is skipped.
    """
    for key, value in source.items():
        if value is not None:
            destination[key] = value



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2021 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

import logging
from abc import ABCMeta
from typing import TYPE_CHECKING, Any, Awaitable, Optional

if TYPE_CHECKING:
    import asyncio
    import queue

from urllib.parse import urljoin

LOGGER = logging.getLogger(__name__)


def batch_endpoint_url(server_address: str) -> str:
    # TODO: Check that server_address finishes with a "/"
    return urljoin(server_address, "events/batch")


def labels_endpoint_url(server_address: str) -> str:
    return urljoin(server_address, "labels/batch")


CLOSE_MESSAGE = object()


class BaseSender(metaclass=ABCMeta):
    # Add sender_queue attribute to satisfy mypy
    sender_queue: Optional["queue.Queue[Any] | asyncio.Queue[Any]"] = None

    def ping_backend(self) -> Optional[Awaitable[Any]]:
        """Allows to check backend status.
        Raise CometMPMBackendException if backend is not healthy.
        Raise CometMPMBackendIsNotAvailable if MPM backend is not installed."""
        ...  # pragma: no cover

    def connect(self) -> Any:
        ...  # pragma: no cover

    def put(self, item: Any) -> Optional[Awaitable[None]]:
        ...  # pragma: no cover

    def close(self, timeout: int) -> None:
        ...  # pragma: no cover

    def join(self, timeout: int) -> Optional[Awaitable[None]]:
        ...  # pragma: no cover



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2021-2025 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

import os
from typing import Optional, Tuple, cast

from ..config import get_config
from ..connection import get_api_http_session
from ..exceptions import CometMPMException
from ..settings_helper import get_comet_base_url
from . import rest_client
from .model import Model


class API:
    """
    Main entry point for interacting with the Comet MPM API.

    Provides high-level methods for working with models and workspaces.

    Args:
        api_key: The Comet API key for authentication
    """

    def __init__(self, api_key: Optional[str] = None) -> None:
        """
        Initialize the Comet MPM API client.

        Args:
            api_key: The Comet API key for authentication
        """
        self._config = get_config()
        api_key = self._get_api_key(api_key)

        base_url = get_comet_base_url(api_key, os.environ.get("COMET_URL_OVERRIDE"))
        requests_session = get_api_http_session(api_key)
        self._client = rest_client.Client(requests_session, base_url)

    def _get_api_key(self, api_key: Optional[str]) -> str:
        """
        Makes sure that the API key is available.
        """
        api_key = api_key or self._config.get("comet.api_key")

        if api_key is None:
            raise CometMPMException(
                "COMET_API_KEY is not defined in environment, and api_key is not given"
            )

        return api_key

    def get_model_by_name(
        self, workspace_name: str, model_name: str
    ) -> Optional[Model]:
        """
        Get a model instance by workspace name and model name.

        Args:
            workspace_name: Name of the workspace containing the model
            model_name: Name of the model to retrieve

        Returns:
            Model: A Model instance for the specified model, if it exists, otherwise None
        """
        model_id = self.get_model_id(workspace_name, model_name)
        if model_id is not None:
            return Model(client=self._client, model_id=model_id)
        else:
            return None

    def get_model(self) -> Optional[Model]:
        """
        Get the default model for this dashboard in a Python Panel environment.

        Returns:
            Model: A Model instance for the model
        """
        model_id = self._config.get_option("modelId")
        if model_id:
            return Model(client=self._client, model_id=model_id)
        else:
            return None

    def get_model_id(self, workspace_name: str, model_name: str) -> Optional[str]:
        """
        Get the model ID for a given workspace and model name.

        Args:
            workspace_name: Name of the workspace containing the model
            model_name: Name of the model to retrieve

        Returns:
            Optional[str]: The model ID if found, None otherwise
        """
        return self._client.get_model_id(workspace_name, model_name)

    def get_panel_workspace(self) -> Optional[str]:
        """
        Get the workspace name from panel options.

        Returns:
            Optional[str]: The workspace name from panel options, or None if not found
        """
        return cast(Optional[str], self._config.get("comet.workspace"))

    def get_panel_width(self) -> int:
        """
        Get the panel width from configuration.

        Returns:
            int: The panel width
        """
        return cast(int, self._config.get("COMET_PANEL_WIDTH"))

    def get_panel_height(self) -> int:
        """
        Get the panel height from configuration.

        Returns:
            int: The panel height
        """
        return cast(int, self._config.get("COMET_PANEL_HEIGHT"))

    def get_panel_size(self) -> Tuple[int, int]:
        """
        Get the panel size (width, height) from configuration.

        Returns:
            Tuple[int, int]: The panel size as (width, height)
        """
        width: int = cast(int, self._config.get("COMET_PANEL_WIDTH"))
        height: int = cast(int, self._config.get("COMET_PANEL_HEIGHT"))
        return (width, height)



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.com
#  Copyright (C) 2021-2024 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************
import os

AWS_LAMBDA_ENVIRONMENT_MARKER_KEY = "LAMBDA_TASK_ROOT"


def is_aws_lambda_environment() -> bool:
    """Allows to check if we are executing in AWS lambda environment"""
    return AWS_LAMBDA_ENVIRONMENT_MARKER_KEY in os.environ



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2021 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************
from comet_mpm.logging_messages import (
    ASYNCIO_SENDER_NOT_CONNECTED_ERROR,
    BACKEND_IS_NOT_AVAILABLE_ERROR,
)


class CometMPMException(Exception):
    """A common Exception class, all exceptions raised by comet_mpm will be children of this class"""


class AsyncioSenderNoConnected(CometMPMException):
    def __init__(self) -> None:
        self.args = (ASYNCIO_SENDER_NOT_CONNECTED_ERROR,)


class CometMPMBackendException(CometMPMException):
    def __init__(self, message: str):
        super().__init__(message)


class CometMPMBackendIsNotAvailable(CometMPMBackendException):
    def __init__(self) -> None:
        super().__init__(BACKEND_IS_NOT_AVAILABLE_ERROR)



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2021 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

"""
Event sending utilities for Comet MPM.

This package provides sender classes for asynchronously sending events
to the Comet MPM backend, including both thread-based and asyncio-based implementations.
"""

import logging
import sys
from typing import TYPE_CHECKING, Optional

from ..logging_messages import GEVENT_NOT_SUPPORTED
from ..server_address import ServerAddress
from .asyncio_sender import get_asyncio_sender
from .base import BaseSender
from .thread_sender import get_thread_sender

if TYPE_CHECKING:
    from .._logging import ErrorStore

LOGGER = logging.getLogger(__name__)


def get_sender(
    api_key: str,
    server_address: ServerAddress,
    max_batch_size: int,
    max_batch_time: int,
    batch_sending_timeout: int,
    asyncio: bool = False,
    error_store: Optional["ErrorStore"] = None,
) -> "BaseSender":

    if "gevent" in sys.modules:
        import gevent.monkey

        # Inspired by gevent.monkey.is_anything_patched
        is_gevent_active = bool(gevent.monkey.saved)

        if is_gevent_active:
            LOGGER.warning(GEVENT_NOT_SUPPORTED)

    if asyncio is True:
        return get_asyncio_sender(
            api_key=api_key,
            server_address=server_address,
            max_batch_size=max_batch_size,
            max_batch_time=max_batch_time,
            batch_sending_timeout=batch_sending_timeout,
            error_store=error_store,
        )
    else:
        return get_thread_sender(
            api_key=api_key,
            server_address=server_address,
            max_batch_size=max_batch_size,
            max_batch_time=max_batch_time,
            batch_sending_timeout=batch_sending_timeout,
            error_store=error_store,
        )



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2024 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************
from urllib.parse import urljoin

from .connection_helpers import sanitize_url


class ServerAddress:
    """
    Encapsulates information about server API address and endpoints
    Args:
        base_url: the base URL (https://ww.comet.com/mpm/)
        api_version: the current API version (v2)
    """

    def __init__(self, base_url: str, api_version: str):
        self.base_url = base_url
        self.api_version = api_version
        self._api_full_path = sanitize_url(urljoin(base_url, api_version))

    @property
    def api_path(self) -> str:
        return self._api_full_path

    def __str__(self) -> str:
        return self.api_path

    def __repr__(self) -> str:
        return self.api_path

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ServerAddress):
            return NotImplemented
        return self.api_path == other.api_path

    def batch_endpoint_url(self) -> str:
        return urljoin(self._api_full_path, "events/batch")

    def labels_endpoint_url(self) -> str:
        return urljoin(self._api_full_path, "labels/batch")

    def is_alive_endpoint_url(self) -> str:
        return urljoin(self.base_url, "isAlive/ping")



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2021 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

BACKEND_IS_NOT_AVAILABLE_ERROR = "Comet Production Monitoring don't seems to be installed in your installation. Please contact your Comet Admin."

BACKEND_500_ERROR = "Comet Production Monitoring isn't working as expected, Please contact your Comet admin."

BATCH_SENDING_ERROR = "Error sending batch data to the backend"

UPLOAD_CSV_DATASET_ERROR = "Error on uploading csv dataset"

BATCH_SENDING_EXCEPTION = "Unknown exception happened when sending data to the backend"

LABEL_EVENT_SEND_ERROR = "Error sending label to the backend"

GEVENT_NOT_SUPPORTED = (
    "Using the MPM SDK with gevent is not officially supported, data loss might occurs"
)

ASYNCIO_SENDER_NOT_CONNECTED_ERROR = (
    "You need to call CometMPM.connect before sending events with CometMPM.log_event"
)

ASYNCIO_SENDER_HARD_SHUTDOWN = "There are still some data in-flight and might be missing. Make sure to call CometMPM.join() manually to guarantee that all data has been sent to the Backend"

ERROR_QUANTILES_LIST_EMPTY = (
    "Error cannot compute distribution summary because the quantiles list is empty."
)
ERROR_QUANTILES_INVALID_VALUE = (
    "The quantile must be within [0,1]. The current min value %f and max %f"
)

LOGGING_DIRECTORY_CREATION_ERROR = "Cannot create path to log file %r"

LOGGING_LOG_FILE_OPEN_ERROR = "Cannot open log file %r; file logging is disabled"

INVALID_SETTING_DEFAULT_FALLBACK = "Some MPM settings values were invalid, default values were used instead, see below for the list"

UNEXPECTED_EVENT_TYPE = "Unexpected event type %r"

DEPRECATED_OUTPUT_VALUE_AND_PROBABILITY_WITHOUT_FEATURES = 'You are using the deprecated fields `output_value` and `output_probability`,you should specify `output_features={"value": "<>", "probability": "<>"}`'

DEPRECATED_OUTPUT_VALUE_AND_PROBABILITY_WITH_FEATURES = 'You are using the deprecated fields `output_value` and `output_probability`, you should specify `output_features={"value": "<>", "probability": "<>", ...}`'

DEPRECATED_DATAFRAME_OUTPUT_VALUE_AND_PROBABILITY = "You are using the deprecated fields `output_value_column` and `output_probability_column`,you should specify `output_features_columns with value and probability columns`"

MPM_ALREADY_CLOSED_LOG_DATA_WARNING = "This MPM instance has already been closed. Create a new instance to log additional data."

MPM_IN_AWS_LAMBDA_NEEDS_END = "As you are running in a Lambda function, you will need to call 'mpm.end()' when finished to ensure all data is logged before exiting."

MPM_JOIN_DEPRECATED_WARNING = "MPM.join is deprecated, use MPM.end instead."

MPM_LABEL_DEPRECATED_WARNING = (
    "label parameter deprecated, use labels parameter instead."
)

PARSE_API_KEY_EMPTY_KEY = "Can not parse empty Comet API key"

PARSE_API_KEY_EMPTY_EXPECTED_ATTRIBUTES = (
    "Expected attributes not found in the Comet API key: %r"
)

PARSE_API_KEY_TOO_MANY_PARTS = "Too many parts (%d) found in the Comet API key: %r"

BASE_URL_MISMATCH_CONFIG_API_KEY = "Comet URL conflict detected between config (%r) and API Key (%r). MPM SDK will use config URL. Resolve by either removing config URL or set it to the same value."
